import numpy as np
from keras_cv_attention_models import backend
from keras_cv_attention_models.backend import layers, models, functional, image_data_format, initializers

""" Convert and replacing """


class SAMModel(models.Model):
    """
    Arxiv article: [Sharpness-Aware Minimization for Efficiently Improving Generalization](https://arxiv.org/pdf/2010.01412.pdf)
    Implementation by: [Keras SAM (Sharpness-Aware Minimization)](https://qiita.com/T-STAR/items/8c3afe3a116a8fc08429)

    Usage is same with `keras.models.Model`: `model = SAMModel(inputs, outputs, rho=sam_rho, name=name)`
    """

    def __init__(self, *args, rho=0.05, **kwargs):
        super().__init__(*args, **kwargs)
        import tensorflow as tf

        self.rho = tf.constant(rho, dtype=tf.float32)
        self.tf = tf

    def train_step(self, data):
        if len(data) == 3:
            x, y, sample_weight = data
        else:
            sample_weight = None
            x, y = data

        # 1st step
        with self.tf.GradientTape() as tape:
            y_pred = self(x, training=True)
            loss = self.compiled_loss(y, y_pred, sample_weight=sample_weight, regularization_losses=self.losses)

        trainable_vars = self.trainable_variables
        gradients = tape.gradient(loss, trainable_vars)

        norm = self.tf.linalg.global_norm(gradients)
        scale = self.rho / (norm + 1e-12)
        e_w_list = []
        for v, grad in zip(trainable_vars, gradients):
            e_w = grad * scale
            v.assign_add(e_w)
            e_w_list.append(e_w)

        # 2nd step
        with self.tf.GradientTape() as tape:
            y_pred_adv = self(x, training=True)
            loss_adv = self.compiled_loss(y, y_pred_adv, sample_weight=sample_weight, regularization_losses=self.losses)
        gradients_adv = tape.gradient(loss_adv, trainable_vars)
        for v, e_w in zip(trainable_vars, e_w_list):
            v.assign_sub(e_w)

        # optimize
        self.optimizer.apply_gradients(zip(gradients_adv, trainable_vars))

        self.compiled_metrics.update_state(y, y_pred, sample_weight=sample_weight)
        return_metrics = {}
        for metric in self.metrics:
            result = metric.result()
            if isinstance(result, dict):
                return_metrics.update(result)
            else:
                return_metrics[metric.name] = result
        return return_metrics


@backend.register_keras_serializable(package="model_surgery")
class DropConnect(layers.Layer):
    def __init__(self, rate=0, **kwargs):
        super(DropConnect, self).__init__(**kwargs)
        self.rate = rate
        self.supports_masking = False

    def build(self, input_shape):
        if self.rate > 0:
            noise_shape = [None] + [1] * (len(input_shape) - 1)  # [None, 1, 1, 1]
            self.drop = layers.Dropout(self.rate, noise_shape=noise_shape, name=self.name + "drop")
        else:
            self.drop = lambda xx: xx

    def call(self, inputs, **kwargs):
        shortcut, deep = inputs
        deep = self.drop(deep)
        return layers.Add()([shortcut, deep])

    def get_config(self):
        config = super(DropConnect, self).get_config()
        config.update({"rate": self.rate})
        return config


def add_l2_regularizer_2_model(model, weight_decay, custom_objects={}, apply_to_batch_normal=False, apply_to_bias=False):
    # https://github.com/keras-team/keras/issues/2717#issuecomment-456254176
    from tensorflow import keras

    if 0:
        regularizers_type = {}
        for layer in model.layers:
            rrs = [kk for kk in layer.__dict__.keys() if "regularizer" in kk and not kk.startswith("_")]
            if len(rrs) != 0:
                # print(layer.name, layer.__class__.__name__, rrs)
                if layer.__class__.__name__ not in regularizers_type:
                    regularizers_type[layer.__class__.__name__] = rrs
        print(regularizers_type)

    for layer in model.layers:
        attrs = []
        if isinstance(layer, layers.Dense) or isinstance(layer, layers.Conv2D):
            # print(">>>> Dense or Conv2D", layer.name, "use_bias:", layer.use_bias)
            attrs = ["kernel_regularizer"]
            if apply_to_bias and layer.use_bias:
                attrs.append("bias_regularizer")
        elif isinstance(layer, layers.DepthwiseConv2D):
            # print(">>>> DepthwiseConv2D", layer.name, "use_bias:", layer.use_bias)
            attrs = ["depthwise_regularizer"]
            if apply_to_bias and layer.use_bias:
                attrs.append("bias_regularizer")
        elif isinstance(layer, layers.SeparableConv2D):
            # print(">>>> SeparableConv2D", layer.name, "use_bias:", layer.use_bias)
            attrs = ["pointwise_regularizer", "depthwise_regularizer"]
            if apply_to_bias and layer.use_bias:
                attrs.append("bias_regularizer")
        elif apply_to_batch_normal and isinstance(layer, layers.BatchNormalization):
            # print(">>>> BatchNormalization", layer.name, "scale:", layer.scale, ", center:", layer.center)
            if layer.center:
                attrs.append("beta_regularizer")
            if layer.scale:
                attrs.append("gamma_regularizer")
        elif apply_to_batch_normal and isinstance(layer, layers.PReLU):
            # print(">>>> PReLU", layer.name)
            attrs = ["alpha_regularizer"]

        for attr in attrs:
            if hasattr(layer, attr) and layer.trainable:
                setattr(layer, attr, keras.regularizers.L2(weight_decay / 2))

    # So far, the regularizers only exist in the model config. We need to
    # reload the model so that Keras adds them to each layer's losses.
    # temp_weight_file = "tmp_weights.h5"
    # model.save_weights(temp_weight_file)
    # out_model = models.model_from_json(model.to_json(), custom_objects=custom_objects)
    # out_model.load_weights(temp_weight_file, by_name=True)
    # os.remove(temp_weight_file)
    # return out_model
    return models.clone_model(model)


def replace_ReLU(model, target_activation="PReLU", **kwargs):
    from tensorflow import keras
    from tensorflow.keras.layers import ReLU, PReLU, Activation

    def convert_ReLU(layer):
        # print(layer.name)
        if isinstance(layer, ReLU) or (isinstance(layer, Activation) and layer.activation == keras.activations.relu):
            if target_activation == "PReLU":
                layer_name = layer.name.replace("_relu", "_prelu")
                print(">>>> Convert ReLU:", layer.name, "-->", layer_name)
                # Default initial value in mxnet and pytorch is 0.25
                return PReLU(shared_axes=[1, 2], alpha_initializer=initializers.Constant(0.25), name=layer_name, **kwargs)
            elif isinstance(target_activation, str):
                layer_name = layer.name.replace("_relu", "_" + target_activation)
                print(">>>> Convert ReLU:", layer.name, "-->", layer_name)
                return Activation(activation=target_activation, name=layer_name, **kwargs)
            else:
                act_class_name = target_activation.__name__
                layer_name = layer.name.replace("_relu", "_" + act_class_name)
                print(">>>> Convert ReLU:", layer.name, "-->", layer_name)
                return target_activation(**kwargs)
        return layer

    input_tensors = layers.Input(model.input_shape[1:])
    return models.clone_model(model, input_tensors=input_tensors, clone_function=convert_ReLU)


def change_model_input_shape(model, new_input_shape):
    import json
    import os

    new_input_shape = [None, *new_input_shape[: len(model.input_shape) - 2], model.input_shape[-1]]
    if list(model.input_shape) == new_input_shape:
        return model

    aa = json.loads(model.to_json())
    aa["config"]["layers"][0]["config"]["batch_input_shape"] = new_input_shape
    bb = models.model_from_json(json.dumps(aa))
    temp_name = "__change_model_input_shape_temp__.h5"
    model.save_weights(temp_name)
    bb.load_weights(temp_name)
    os.remove(temp_name)
    print(">>>> Changed model input shape from {} to {}".format(model.input_shape, bb.input_shape))
    return bb


def convert_to_dynamic_input_shape(model):
    if not backend.is_tensorflow_backend:
        return model  # Do nothing for pytorch model

    def __convert_to_dynamic_input_shape__(layer):
        if hasattr(layer, "load_resized_weights"):
            aa = layer.__class__.from_config(layer.get_config())
            # aa.build(layer.input_shape)  # [???] Error when saving ValueError: Unable to create dataset (name already exists)
            _ = aa(initializers.zeros()([1, *layer.input_shape[1:]]))
            aa.set_weights(layer.get_weights())
            return aa
        else:
            return layer

    input_shape = [None] * (len(model.input_shape) - 2) + list(model.input_shape)[-1:]
    bb = models.clone_model(model, layers.Input(input_shape), clone_function=__convert_to_dynamic_input_shape__)
    print(">>>> Changed model input shape from {} to {}".format(model.input_shape, bb.input_shape))
    return bb


# Not using
def find_layer_name_nested(layers, layer_name_cond, reverse=False):
    layers = layers[::-1] if reverse else layers
    for ii in layers:
        if layer_name_cond(ii.name):
            return ii
        if hasattr(ii, "layers"):  # nested
            found_layer = find_layer_name_nested(ii.layers, layer_name_cond, reverse=reverse)
            if found_layer is not None:
                return ii  # Return nested block
    return None


def split_model_to_head_body_tail_by_blocks(model):
    for ii in model.layers:
        if "stack" in ii.name or "block" in ii.name:
            body_start_layer = ii
            break
    assert body_start_layer is not None, "None layer with name containing 'stack' or 'block' found"

    for ii in model.layers[::-1]:
        if "stack" in ii.name or "block" in ii.name:
            body_end_layer = ii
            break
    print(">>>> Split model to head + body + tail by blocks, body_start_layer: {}, body_end_layer: {}".format(body_start_layer.name, body_end_layer.name))

    head_model = models.Model(model.inputs[0], body_start_layer.input, name=model.name + "_head")
    body_model = models.Model(body_start_layer.input, body_end_layer.output, name=model.name + "_body")
    tail_model = models.Model(body_end_layer.output, model.outputs[0], name=model.name + "_tail")
    return head_model, body_model, tail_model


def replace_add_with_stochastic_depth(model, survivals=(1, 0.8)):
    """
    - [Deep Networks with Stochastic Depth](https://arxiv.org/pdf/1603.09382.pdf)
    - [tfa.layers.StochasticDepth](https://www.tensorflow.org/addons/api_docs/python/tfa/layers/StochasticDepth)
    """
    from tensorflow_addons.layers import StochasticDepth

    add_layers = [ii.name for ii in model.layers if isinstance(ii, layers.Add)]
    total_adds = len(add_layers)
    if isinstance(survivals, float):
        survivals = [survivals] * total_adds
    elif isinstance(survivals, (list, tuple)) and len(survivals) == 2:
        start, end = survivals
        survivals = [start - (1 - end) * float(ii) / total_adds for ii in range(total_adds)]
    survivals_dict = dict(zip(add_layers, survivals))

    def __replace_add_with_stochastic_depth__(layer):
        if isinstance(layer, layers.Add):
            layer_name = layer.name
            new_layer_name = layer_name.replace("_add", "_stochastic_depth")
            new_layer_name = new_layer_name.replace("add_", "stochastic_depth_")
            survival_probability = survivals_dict[layer_name]
            if survival_probability < 1:
                print("Converting:", layer_name, "-->", new_layer_name, ", survival_probability:", survival_probability)
                return StochasticDepth(survival_probability, name=new_layer_name)
            else:
                return layer
        return layer

    input_tensors = layers.Input(model.input_shape[1:])
    return models.clone_model(model, input_tensors=input_tensors, clone_function=__replace_add_with_stochastic_depth__)


def replace_add_with_drop_connect(model, drop_rate=(0, 0.2)):
    """
    - [Deep Networks with Stochastic Depth](https://arxiv.org/pdf/1603.09382.pdf)
    - [tfa.layers.StochasticDepth](https://www.tensorflow.org/addons/api_docs/python/tfa/layers/StochasticDepth)
    """

    add_layers = [ii.name for ii in model.layers if isinstance(ii, layers.Add)]
    total_adds = len(add_layers)
    if isinstance(drop_rate, float):
        drop_rates = [drop_rate] * total_adds
    elif isinstance(drop_rate, (list, tuple)) and len(drop_rate) == 2:
        start, end = drop_rate
        drop_rates = [(end - start) * float(ii) / total_adds for ii in range(total_adds)]
    drop_conn_rate_dict = dict(zip(add_layers, drop_rates))

    def __replace_add_with_stochastic_depth__(layer):
        if isinstance(layer, layers.Add):
            layer_name = layer.name
            new_layer_name = layer_name.replace("_add", "_drop_conn")
            new_layer_name = new_layer_name.replace("add_", "drop_conn_")
            drop_conn_rate = drop_conn_rate_dict[layer_name]
            if drop_conn_rate < 1:
                print("Converting:", layer_name, "-->", new_layer_name, ", drop_conn_rate:", drop_conn_rate)
                return DropConnect(drop_conn_rate, name=new_layer_name)
            else:
                return layer
        return layer

    input_tensors = layers.Input(model.input_shape[1:])
    return models.clone_model(model, input_tensors=input_tensors, clone_function=__replace_add_with_stochastic_depth__)


def replace_stochastic_depth_with_add(model, drop_survival=False):
    from tensorflow_addons.layers import StochasticDepth

    def __replace_stochastic_depth_with_add__(layer):
        if isinstance(layer, StochasticDepth):
            layer_name = layer.name
            new_layer_name = layer_name.replace("_stochastic_depth", "_lambda")
            survival = layer.survival_probability
            print("Converting:", layer_name, "-->", new_layer_name, ", survival_probability:", survival)
            if drop_survival or not survival < 1:
                return layers.Add(name=new_layer_name)
            else:
                return layers.Lambda(lambda xx: xx[0] + xx[1] * survival, name=new_layer_name)
        return layer

    input_tensors = layers.Input(model.input_shape[1:])
    return models.clone_model(model, input_tensors=input_tensors, clone_function=__replace_stochastic_depth_with_add__)


def convert_to_token_label_model(model, pool_layer_id="auto"):
    # Search pool layer id
    num_total_layers = len(model.layers)
    if pool_layer_id == "auto":
        for header_layer_id, layer in enumerate(model.layers[::-1]):
            header_layer_id = num_total_layers - header_layer_id - 1
            print("[Search pool layer] header_layer_id = {}, layer.name = {}".format(header_layer_id, layer.name))
            if isinstance(layer, layers.GlobalAveragePooling2D):
                break
        pool_layer_id = header_layer_id

    nn = model.layers[pool_layer_id - 1].output  # layer output before pool layer

    # Add header layers w/o pool layer
    for header_layer_id in range(pool_layer_id + 1, num_total_layers):
        aa = model.layers[header_layer_id]
        config = aa.get_config()
        config["name"] = config["name"] + "_token_label"
        if isinstance(aa, layers.LayerNormalization) and config["axis"] == [1]:
            config["axis"] = [-1]
        # print(config)
        print("[Build new layer] header_layer_id = {}, layer.name = {}".format(header_layer_id, config["name"]))

        bb = aa.__class__.from_config(config)
        bb.build(nn.shape)
        bb.set_weights(aa.get_weights())
        nn = bb(nn)
    token_label_model = models.Model(model.inputs[0], [*model.outputs, nn])
    print("token_label_model.output_shape =", token_label_model.output_shape)
    return token_label_model


""" Get model info """


def get_actual_survival_probabilities(model):
    from tensorflow_addons.layers import StochasticDepth

    return [ii.survival_probability for ii in model.layers if isinstance(ii, StochasticDepth)]


def get_actual_drop_connect_rates(model):
    return [ii.rate for ii in model.layers if isinstance(ii, layers.Dropout) or isinstance(ii, DropConnect)]


def get_pyramide_feature_layers(model, match_reg="^stack_?(\\d+).*output.*$"):
    """Pick all stack output layers"""
    import re

    if hasattr(model, "extract_features"):
        return model.extract_features()

    dd = {}
    pre_stack, pre_output_shape = 0, model.input_shape[1:]
    for ii in model.layers:
        cur_name = ii.name
        # if cur_name.startswith("pre_output") and ii.output_shape[1:] == pre_output_shape:
        #     cur_name = "stack_{}".format(pre_stack) + cur_name  # For Swin

        matched = re.match(match_reg, cur_name)
        if matched is not None:
            cur_stack = "stack_" + matched[1] + "_output"
            pre_stack, pre_output_shape = matched[1], ii.output_shape[1:]
            dd.update({cur_stack: ii})

    """ Filter those have same downsample rate """
    ee = {str(vv.output_shape[2]): vv for kk, vv in dd.items()}
    return list(ee.values())


def align_pyramide_feature_output_by_image_data_format(pyramide_feature_layers):
    aa = [ii.output_shape for ii in pyramide_feature_layers]
    bb = [ii[1] for ii in aa]
    if all([ii is None or bb[id - 1] is None or ii <= bb[id - 1] for id, ii in enumerate(bb[1:], start=1)]):
        features_data_format = "channels_last"
    else:
        features_data_format = "channels_first"

    if features_data_format == backend.image_data_format():
        output_names = [ii.name for ii in pyramide_feature_layers]
        outputs = [ii.output for ii in pyramide_feature_layers]
    elif features_data_format == "channels_last":  # backend.image_data_format is "channels_first"
        output_names = [ii.name + "_perm" for ii in pyramide_feature_layers]
        outputs = [layers.Permute([3, 1, 2], name=ii.name + "_perm")(ii.output) for ii in pyramide_feature_layers]
    else:  # features_data_format is "channels_first" and backend.image_data_format is "channels_last". Should not happen
        output_names = [ii.name + "_perm" for ii in pyramide_feature_layers]
        outputs = [layers.Permute([2, 3, 1], name=ii.name + "_perm")(ii.output) for ii in pyramide_feature_layers]
    return output_names, outputs


def get_global_avg_pool_layer_id(model):
    """Search GlobalAveragePooling2D layer id"""
    num_total_layers = len(model.layers)
    for header_layer_id, layer in enumerate(model.layers[::-1]):
        header_layer_id = num_total_layers - header_layer_id - 1
        print("[Search pool layer] header_layer_id = {}, layer.name = {}".format(header_layer_id, layer.name))
        if isinstance(layer, layers.GlobalAveragePooling2D):
            break
    return header_layer_id


def get_layer_as_inputs_count_from_mode_json(model):
    # dd = {}
    # for ii in cc.layers:
    #     layer_inputs = ii.input if isinstance(ii.input, (list, tuple)) else [ii.input]
    #     for layer_input in layer_inputs:
    #         dd[layer_input.node.layer.name] = dd.get(layer_input.node.layer.name, 0) + 1
    if isinstance(model, dict):
        model_config = model
    else:
        import json

        model_config = json.loads(model.to_json())

    dd = {}
    is_inbound_elem = lambda xx: isinstance(xx, list) and isinstance(xx[0], str)
    for layer in model_config["config"]["layers"]:
        # print(layer['inbound_nodes'])
        for ii in layer["inbound_nodes"]:
            # print(ii)
            # single input layer: [[['stack4_block2_1_conv', 0, 0, {}]]]
            # multi input layer: 'inbound_nodes': [['stack4_block2_1_conv', 0, 0, {'y': ['stack4_block2_1_short', 0, 0], 'name': None}]]}
            if is_inbound_elem(ii):
                dd[ii[0]] = dd.get(ii[0], 0) + 1
                for kk, vv in ii[3].items():
                    if is_inbound_elem(vv):
                        dd[vv[0]] = dd.get(vv[0], 0) + 1
            elif isinstance(ii, list) and isinstance(ii[0], list):
                for jj in ii:
                    dd[jj[0]] = dd.get(jj[0], 0) + 1
                    for kk, vv in jj[3].items():
                        if is_inbound_elem(vv):
                            dd[vv[0]] = dd.get(vv[0], 0) + 1
    return dd


def get_flops(model, input_shape=None):
    input_shape = [(1, *ii.shape[1:]) for ii in model.inputs] if input_shape is None else input_shape
    if backend.is_tensorflow_backend:
        # https://github.com/tensorflow/tensorflow/issues/32809#issuecomment-849439287
        import tensorflow as tf
        from tensorflow.python.profiler import model_analyzer, option_builder

        input_signature = [tf.TensorSpec(shape=ss, dtype=ii.dtype, name=ii.name) for ii, ss in zip(model.inputs, input_shape)]
        forward_graph = tf.function(model, [input_signature]).get_concrete_function().graph
        options = option_builder.ProfileOptionBuilder.float_operation()
        graph_info = model_analyzer.profile(forward_graph, options=options)
        flops = graph_info.total_float_ops // 2
    else:
        import thop
        import torch

        inputs = [torch.ones(ii) for ii in input_shape]
        flops, params = thop.profile(model, inputs=tuple([inputs]))
    print(">>>> FLOPs: {:,}, GFLOPs: {:.4f}G".format(flops, flops / 1e9))
    return flops


def count_params(model):
    # "parameters" and "requires_grad" are for torch model, while "weights" and "trainable" for TF models
    total_params, trainable_params = 0, 0
    for ii in model.parameters() if hasattr(model, "parameters") else model.weights:
        cur_params = np.prod(ii.shape)
        total_params += cur_params
        trainable = ii.requires_grad if hasattr(ii, "requires_grad") else ii.trainable
        trainable_params += cur_params if trainable else 0
    non_trainable_params = total_params - trainable_params
    print("Total params: {:,} | Trainable params: {:,} | Non-trainable params:{:,}".format(total_params, trainable_params, non_trainable_params))
    return total_params, trainable_params


""" Inference and deploy """


def convert_to_mixed_float16(model, convert_batch_norm=False, policy_name="mixed_float16"):
    from tensorflow import keras
    from tensorflow.keras.layers import InputLayer, Activation
    from tensorflow.keras.activations import linear

    policy = keras.mixed_precision.Policy(policy_name)
    policy_config = keras.utils.serialize_keras_object(policy)

    def do_convert_to_mixed_float16(layer):
        if not convert_batch_norm and isinstance(layer, layers.BatchNormalization):
            return layer
        if not isinstance(layer, InputLayer) and not (isinstance(layer, Activation) and layer.activation == linear):
            aa = layer.get_config()
            aa.update({"dtype": policy_config})
            bb = layer.__class__.from_config(aa)
            bb.build(layer.input_shape)
            bb.set_weights(layer.get_weights())
            return bb
        return layer

    input_tensors = layers.Input(model.input_shape[1:])
    return models.clone_model(model, input_tensors=input_tensors, clone_function=do_convert_to_mixed_float16)


def convert_mixed_float16_to_float32(model):
    from tensorflow.keras.layers import InputLayer, Activation
    from tensorflow.keras.activations import linear

    def do_convert_to_mixed_float16(layer):
        if not isinstance(layer, InputLayer) and not (isinstance(layer, Activation) and layer.activation == linear):
            aa = layer.get_config()
            aa.update({"dtype": "float32"})
            bb = layer.__class__.from_config(aa)
            bb.build(layer.input_shape)
            bb.set_weights(layer.get_weights())
            return bb
        return layer

    input_tensors = layers.Input(model.input_shape[1:])
    return models.clone_model(model, input_tensors=input_tensors, clone_function=do_convert_to_mixed_float16)


def fuse_layer_single_input(model, fuse_layer_condition, pre_layer_condition=None, layer_config_process=None, layer_weight_process=None, verbose=0):
    import json

    pre_layer_condition = (lambda layer: True) if pre_layer_condition is None else pre_layer_condition

    """ Check bn layers with conv layer input """
    model_config = json.loads(model.to_json())
    ee = {layer["name"]: layer for layer in model_config["config"]["layers"]}
    pre_fused_layers, fused_layers = [], []
    for layer in model_config["config"]["layers"]:
        if fuse_layer_condition(layer) and len(layer["inbound_nodes"]) == 1:
            input_node = layer["inbound_nodes"][0][0]
            if isinstance(input_node, list) and pre_layer_condition(ee.get(input_node[0], {})):
                pre_fused_layers.append(input_node[0])
                fused_layers.append(layer["name"])
    if verbose > 0:
        print(">>>> pre_fused_layers =", pre_fused_layers, "fused_layers =", fused_layers)
        print(">>>> len(pre_fused_layers) =", len(pre_fused_layers), "len(fused_layers) =", len(fused_layers))
        print()
        # len(fuse_convs) = 53, len(fuse_bns) = 53

    """ Create new model config """
    layers = []
    fused_layers_dict = dict(zip(fused_layers, pre_fused_layers))
    pre_fused_layers_dict = dict(zip(pre_fused_layers, fused_layers))
    is_inbound_elem = lambda xx: isinstance(xx, list) and isinstance(xx[0], str)
    for layer in model_config["config"]["layers"]:
        if layer["name"] in pre_fused_layers and layer_config_process is not None:
            if verbose > 0:
                print(">>>> Create new layer config for:", layer["name"])
            fused_layer_config = ee.get(pre_fused_layers_dict[layer["name"]])["config"]
            layer["config"] = layer_config_process(layer["config"], fused_layer_config)
        elif layer["name"] in fused_layers:
            if verbose > 0:
                print(">>>> Remove layer:", layer["name"])
            continue

        for ii in layer["inbound_nodes"]:
            # print(ii)
            # single input layer: [[['stack4_block2_1_conv', 0, 0, {}]]]
            # multi input layer: 'inbound_nodes': [['stack4_block2_1_conv', 0, 0, {'y': ['stack4_block2_1_short', 0, 0], 'name': None}]]}
            # multi_input layer: 'inbound_nodes': [[['stack1_block1_1_conv', 0, 0, {}], ['stem_1_conv', 0, 0, {}], ['stem_2_conv', 0, 0, {}]]]}
            if is_inbound_elem(ii):
                # print(">>>> Replace inbound_nodes: {}, {} --> {}".format(layer["name"], ii[0], fused_bn_dict[ii[0]]))
                ii[0] = fused_layers_dict.get(ii[0], ii[0])
                ii[3] = {kk: [fused_layers_dict.get(vv[0], vv[0]), *vv[1:]] if is_inbound_elem(vv) else vv for kk, vv in ii[3].items()}
            elif isinstance(ii, list) and isinstance(ii[0], list):
                for jj in ii:
                    jj[0] = fused_layers_dict.get(jj[0], jj[0])
                    jj[3] = {kk: [fused_layers_dict.get(vv[0], vv[0]), *vv[1:]] if is_inbound_elem(vv) else vv for kk, vv in jj[3].items()}

        layers.append(layer)
    model_config["config"]["layers"] = layers
    new_model = models.model_from_json(json.dumps(model_config))
    if verbose > 0:
        print()

    """ New model set layer weights by layer names """
    for layer in new_model.layers:
        if layer.name in fused_layers:  # This should not happen
            continue

        orign_layer = model.get_layer(layer.name)
        if layer.name in pre_fused_layers_dict and layer_weight_process is not None:
            orign_fused_layer = model.get_layer(pre_fused_layers_dict[layer.name])
            if verbose > 0:
                print(">>>> Fuse weights {} <- {}".format(layer.name, orign_fused_layer.name))
            layer.set_weights(layer_weight_process(orign_layer, orign_fused_layer))
        else:
            layer.set_weights(orign_layer.get_weights())
    return new_model


def fuse_bn_dense_weights(bn_layer, dense_layer):
    """
    https://github.com/THU-MIG/RepViT/blob/main/model/repvit.py#L168
    Convert model by fusing batchnorm + dense

    Exampls:
    >>> from keras_cv_attention_models.model_surgery import model_surgery
    >>> ii = tf.random.uniform([1, 32, 32, 3])
    >>> mm = keras.models.Sequential([keras.layers.Input([32, 32, 3]), keras.layers.BatchNormalization(), keras.layers.Dense(10)])
    >>> aa = keras.layers.Dense(10)
    >>> aa.build([None, 32, 32, 3])
    >>> aa.set_weights(model_surgery.fuse_bn_dense_weights(mm.layers[0], mm.layers[1]))
    >>> print(f"{np.allclose(mm(ii), aa(ii), atol=1e-5) = }")
    """
    # BatchNormalization returns: gamma * (batch - self.moving_mean) / sqrt(self.moving_var + epsilon) + beta
    # BN -> Dense = (gamma * (batch - self.moving_mean) / sqrt(self.moving_var + epsilon)) @ ww + beta @ ww + bb
    # --> gamma / sqrt(self.moving_var + epsilon) @ ww * batch + beta @ ww + bb - gamma * self.moving_mean / sqrt(self.moving_var + epsilon) @ ww
    # --> new_ww = gamma / sqrt(self.moving_var + epsilon) @ ww
    # --> new_bias = beta @ ww + bb - self.moving_mean @ new_ww
    ww = dense_layer.kernel
    bb = dense_layer.bias if dense_layer.use_bias else 0
    batch_std = functional.sqrt(bn_layer.moving_variance + bn_layer.epsilon)

    # ww: [in, out], bb: [out], beta, gamma, moving_mean, moving_variance: [in]
    # -> new_ww: [in, out], new_bias: [out]
    new_ww = ww * (bn_layer.gamma / batch_std)[:, None]
    new_bias = bn_layer.beta[None, :] @ ww + bb - bn_layer.moving_mean[None, :] @ new_ww
    return [new_ww, new_bias[0]]


def fuse_conv_bn_weights(conv_layer, bn_layer):
    """
    Convert model by fusing Conv + batchnorm

    Exampls:
    >>> from keras_cv_attention_models import model_surgery
    >>> mm = keras.applications.ResNet50()
    >>> bb = model_surgery.convert_to_fused_conv_bn_model(mm)
    """
    # BatchNormalization returns: gamma * (batch - self.moving_mean) / sqrt(self.moving_var + epsilon) + beta
    # --> new_ww = gamma * conv_w / np.sqrt(var + epsilon)
    # --> new_bb = gamma * (conv_b - mean) / sqrt(var + epsilon) + beta
    batch_std = functional.sqrt(bn_layer.moving_variance + bn_layer.epsilon)

    ww = functional.transpose(conv_layer.depthwise_kernel, [0, 1, 3, 2]) if isinstance(conv_layer, layers.DepthwiseConv2D) else conv_layer.kernel
    bb = conv_layer.bias if conv_layer.use_bias else 0

    new_ww = ww * bn_layer.gamma / batch_std
    new_ww = functional.transpose(new_ww, [0, 1, 3, 2]) if isinstance(conv_layer, layers.DepthwiseConv2D) else new_ww
    new_bb = bn_layer.gamma * (bb - bn_layer.moving_mean) / batch_std + bn_layer.beta
    return [new_ww, new_bb]
    # cc = conv_layer.get_config()
    # cc["use_bias"] = True
    # fused_conv_bn = conv_layer.__class__.from_config(cc)
    # fused_conv_bn.build(conv_layer.input_shape)
    # fused_conv_bn.set_weights([ww, bias])
    # return fused_conv_bn


def convert_to_fused_conv_bn_model(model, verbose=0):
    """
    Convert model by fusing Conv + batchnorm

    Exampls:
    >>> from keras_cv_attention_models import model_surgery
    >>> mm = keras.applications.ResNet50()
    >>> bb = model_surgery.convert_to_fused_conv_bn_model(mm)
    """
    fuse_layer_condition = lambda layer: layer["class_name"] == "BatchNormalization"
    pre_layer_condition = lambda layer: layer.get("class_name") in ["Conv2D", "DepthwiseConv2D"]

    def layer_config_process(pre_layer_config, fused_layer_config):
        pre_layer_config["use_bias"] = True
        return pre_layer_config

    layer_weight_process = lambda pre_fused_layer, fused_layer: fuse_conv_bn_weights(pre_fused_layer, fused_layer)

    return fuse_layer_single_input(
        model=model,
        fuse_layer_condition=fuse_layer_condition,
        pre_layer_condition=pre_layer_condition,
        layer_config_process=layer_config_process,
        layer_weight_process=layer_weight_process,
        verbose=verbose,
    )


def remove_layer_single_input(model, remove_layer_condition, pre_layer_condition=None, layer_config_process=None, layer_weight_process=None, verbose=0):
    return fuse_layer_single_input(model, remove_layer_condition, pre_layer_condition, layer_config_process, layer_weight_process, verbose=verbose)


def fuse_sequential_conv_strict(model, verbose=0):
    """
    Convert model by fusing Conv + batchnorm

    Exampls:
    >>> from keras_cv_attention_models import model_surgery
    >>> mm = keras.applications.ResNet50()
    >>> bb = model_surgery.convert_to_fused_conv_bn_model(mm)
    """
    layer_as_inputs_count = get_layer_as_inputs_count_from_mode_json(model)
    fuse_layer_condition = lambda layer: layer["class_name"] == "Conv2D" and max(layer["config"]["kernel_size"]) == 1
    pre_layer_condition = lambda layer: layer.get("class_name") in ["Conv2D"] and layer_as_inputs_count[layer["name"]] == 1  # pre layer has only 1 output

    def layer_config_process(pre_layer_config, fused_layer_config):
        pre_layer_config["filters"] = fused_layer_config["filters"]
        pre_layer_config["kernel_size"] = pre_layer_config["kernel_size"] if max(pre_layer_config["kernel_size"]) > 1 else fused_layer_config["kernel_size"]
        pre_layer_config["strides"] = pre_layer_config["strides"] if max(pre_layer_config["strides"]) > 1 else fused_layer_config["strides"]
        pre_layer_config["use_bias"] = pre_layer_config["use_bias"] or fused_layer_config["use_bias"]
        pre_layer_config["padding"] = pre_layer_config["padding"] if pre_layer_config["padding"].lower() != "valid" else fused_layer_config["padding"]

        return pre_layer_config

    def layer_weight_process(pre_fused_layer, fused_layer):
        pre_ww = pre_fused_layer.get_weights()
        pre_ww, pre_bias = pre_ww if len(pre_ww) == 2 else (pre_ww, None)
        fused_ww = fused_layer.get_weights()
        fused_ww, fused_bias = fused_ww if len(fused_ww) == 2 else (fused_ww, None)

        fused_ww = np.squeeze(fused_ww)
        new_ww = pre_ww @ fused_ww

        if pre_bias is None and fused_bias is None:
            return [new_ww]
        if pre_bias is None:
            return [new_ww, fused_bias]

        pre_bias = pre_bias @ fused_ww
        new_bias = pre_bias if fused_bias is None else (fused_bias + pre_bias)
        return [new_ww, new_bias]

    return fuse_layer_single_input(
        model=model,
        fuse_layer_condition=fuse_layer_condition,
        pre_layer_condition=pre_layer_condition,
        layer_config_process=layer_config_process,
        layer_weight_process=layer_weight_process,
        verbose=verbose,
    )


def fuse_channel_affine_to_conv_dense(model, verbose=0):
    fuse_layer_condition = lambda layer: layer["class_name"].split(">")[-1] == "ChannelAffine"
    pre_layer_condition = lambda layer: layer.get("class_name") in ["Conv2D", "DepthwiseConv2D", "Dense"]

    def layer_weight_process(pre_fused_layer, fused_layer):
        pre_ww = pre_fused_layer.get_weights()
        pre_ww, pre_bias = pre_ww if len(pre_ww) == 2 else (pre_ww, None)
        fused_ww = fused_layer.get_weights()
        fused_ww, fused_bias = fused_ww if len(fused_ww) == 2 else (fused_ww, None)

        fused_ww = np.squeeze(fused_ww)
        new_ww = pre_ww * (fused_ww[:, None] if pre_ww.shape[-1] == 1 else fused_ww)
        if pre_bias is None and fused_bias is None:
            return [new_ww]
        if pre_bias is None:
            return [new_ww, fused_bias]

        pre_bias = pre_bias * fused_ww
        new_bias = pre_bias if fused_bias is None else (fused_bias + pre_bias)
        return [new_ww, new_bias]

    return fuse_layer_single_input(
        model=model,
        fuse_layer_condition=fuse_layer_condition,
        pre_layer_condition=pre_layer_condition,
        layer_weight_process=layer_weight_process,
        verbose=verbose,
    )


def fuse_reparam_blocks(model, output_layer_key="REPARAM_out", verbose=0):
    # model = convert_to_fused_conv_bn_model(model)
    fuse_layers = []
    fused_reparam_weights = {}  # Record fused weights first, then set_weights after new model built
    for block_out_layer in model.layers:
        if not block_out_layer.name.endswith(output_layer_key) or len(block_out_layer.input) < 2:
            continue

        prefix = block_out_layer.name.split(output_layer_key)[0]
        if verbose > 0:
            print(">>>> block_out_layer.name:", block_out_layer.name, "prefix:", prefix)

        # Just regard the first input as the one with largest kernel_size.
        # It should always with use_bias=True, the bias may comes from fused BatchNorm layer.
        remain_layer = model.get_layer(block_out_layer.input[0].node.layer.name)

        orign_weights = remain_layer.get_weights()
        orign_weight, orign_bias = orign_weights if len(orign_weights) == 2 else (orign_weights[0], 0)
        for input_layer in block_out_layer.input[1:]:
            input_layer = model.get_layer(input_layer.node.layer.name)
            if verbose > 0:
                print("     input_layer.name:", input_layer.name)

            if not input_layer.name.startswith(prefix) or isinstance(input_layer, layers.BatchNormalization):  # Identity branch
                # Conv2D with groups == 1 -> np.eye(filters), DepthWiseConv2D -> 1, COnv2D with groups != 1 -> partial 1
                in_channels, out_channels = remain_layer.weights[0].shape[2], remain_layer.weights[0].shape[3]
                branch_weight = np.zeros([1, 1, in_channels, out_channels], dtype="float32")
                filters, groups = (in_channels, in_channels) if isinstance(remain_layer, layers.DepthwiseConv2D) else (out_channels, remain_layer.groups)
                group_interval = filters // groups
                for ii in range(filters):
                    branch_weight[:, :, ii, ii % group_interval] = 1.0
                branch_bias = 0
                cur_kernel_size = [1] * len(remain_layer.kernel_size)
            elif hasattr(input_layer, "kernel_size"):
                branch_weights = input_layer.get_weights()
                branch_weight, branch_bias = branch_weights if len(branch_weights) == 2 else (branch_weights[0], 0)
                cur_kernel_size = input_layer.kernel_size
                fuse_layers.append(input_layer.name)
                if isinstance(input_layer.input.node.layer, layers.ZeroPadding2D):
                    fuse_layers.append(input_layer.input.node.layer.name)

            if isinstance(input_layer, layers.BatchNormalization):
                if isinstance(remain_layer, layers.DepthwiseConv2D):
                    fake_conv = layers.DepthwiseConv2D(kernel_size=1, use_bias=False)
                else:
                    fake_conv = layers.Conv2D(out_channels, kernel_size=1, use_bias=False)
                _ = fake_conv(initializers.ones()([1, *input_layer.input_shape[1:]]))
                fake_conv.set_weights([branch_weight])  # Previously created identity branch weights
                branch_weight, branch_bias = fuse_conv_bn_weights(fake_conv, input_layer)
                fuse_layers.append(input_layer.name)

            pad = [(remain - cur) // 2 for remain, cur in zip(remain_layer.kernel_size, cur_kernel_size)]  # (3, 3) -> (1, 1)
            branch_weight = functional.pad(branch_weight, [[pad[0], pad[0]], [pad[1], pad[1]], [0, 0], [0, 0]])
            orign_weight = (orign_weight - branch_weight) if isinstance(block_out_layer, layers.Subtract) else (orign_weight + branch_weight)
            orign_bias = (orign_bias - branch_bias) if isinstance(block_out_layer, layers.Subtract) else (orign_bias + branch_bias)
        fuse_layers.append(block_out_layer.name)
        fused_reparam_weights[remain_layer.name] = [orign_weight, orign_bias]

    cc = remove_layer_single_input(model, remove_layer_condition=lambda layer: layer["name"] in fuse_layers, verbose=verbose)
    for ii in cc.layers:
        if ii.name in fused_reparam_weights:
            if verbose > 0:
                print(">>>> set_weights layer.name:", ii.name)
            ii.set_weights(fused_reparam_weights[ii.name])
    return cc


def fuse_distill_head(model, head="head", distill_head="distill_head", head_bn="head_bn", distill_head_bn="distill_head_bn"):
    head = model.get_layer(head)
    if head_bn is not None:
        head_bn = model.get_layer(head_bn)
        head_ww, head_bb = fuse_bn_dense_weights(head_bn, head)
    else:
        head_ww, head_bb = head.get_weights()

    if distill_head in model.output_names:
        distill_head = model.get_layer(distill_head)
        if distill_head_bn is not None:
            distill_head_bn = model.get_layer(distill_head_bn)
            distill_ww, distill_bb = fuse_bn_dense_weights(distill_head_bn, distill_head)
        else:
            distill_ww, distill_bb = distill_head.get_weights()
        head_ww, head_bb = (distill_ww + head_ww) / 2, (distill_bb + head_bb) / 2

    if head_bn is not None:
        model = remove_layer_single_input(model, remove_layer_condition=lambda layer: layer["name"] in [head_bn.name])
    model = models.Model(model.inputs, model.get_layer(head.name).output)
    model.get_layer(head.name).set_weights([head_ww, head_bb])
    return model


def convert_layers_to_deploy_inplace(model):
    for ii in model.layers:
        if hasattr(ii, "switch_to_deploy"):
            ii.switch_to_deploy()
    return model


""" TFLite and ONNX """


@backend.register_keras_serializable(package="model_surgery")
class SplitConv2D(layers.Conv2D):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.super_class = layers.Conv2D

    def build(self, input_shape):
        import tensorflow as tf

        cc = self.get_config().copy()
        cc.update({"groups": 1, "filters": self.filters // self.groups})
        grouped_input_shape = (*input_shape[:-1], input_shape[-1] // self.groups)
        self.convs = []
        for ii in range(self.groups):
            name_scope = self.name + "_{}".format(ii)
            with tf.name_scope(name_scope) as scope:
                cc["name"] = self.name + "_{}".format(ii)
                conv = self.super_class.from_config(cc)
                conv.build(grouped_input_shape)
            self.convs.append(conv)

    def call(self, inputs, **kwargs):
        return functional.concat([conv(ii) for conv, ii in zip(self.convs, functional.split(inputs, self.groups, axis=-1))], axis=-1)


@backend.register_keras_serializable(package="model_surgery")
class SplitScaledStandardizedConv2D(SplitConv2D):
    def __init__(self, gamma=1.0, eps=1e-5, **kwargs):
        from keras_cv_attention_models import attention_layers

        super().__init__(**kwargs)
        self.super_class = attention_layers.ScaledStandardizedConv2D
        self.eps, self.gamma = eps, gamma

    def get_config(self):
        base_config = super().get_config()
        base_config.update({"eps": self.eps, "gamma": self.gamma})
        return base_config


def convert_groups_conv2d_2_split_conv2d(model):
    def __convert_groups_conv2d_2_split_conv2d__(layer):
        if isinstance(layer, layers.Conv2D) and not isinstance(layer, SplitConv2D) and layer.groups != 1:
            aa = layer.get_config()
            # Check if ScaledStandardizedConv2D or typical Conv2D
            bb = SplitScaledStandardizedConv2D.from_config(aa) if hasattr(layer, "gain") else SplitConv2D.from_config(aa)
            # bb.build(layer.input_shape)   # looks like build not working [ ??? ]
            bb(initializers.ones()([1, *layer.input_shape[1:]]))
            wws = functional.split(layer.get_weights()[0], bb.groups, axis=-1)
            if bb.use_bias:
                bbs = functional.split(layer.get_weights()[1], bb.groups, axis=-1)
            if hasattr(layer, "gain"):
                # ScaledStandardizedConv2D with gain from NFNets
                ggs = functional.split(layer.get_weights()[-1], bb.groups, axis=-1)
            for id in range(bb.groups):
                sub_weights = [wws[id].numpy()]
                if bb.use_bias:
                    sub_weights.append(bbs[id].numpy())
                if hasattr(layer, "gain"):
                    sub_weights.append(ggs[id].numpy())
                bb.convs[id].set_weights(sub_weights)
            return bb
        return layer

    input_tensors = layers.Input(model.input_shape[1:])
    return models.clone_model(model, input_tensors=input_tensors, clone_function=__convert_groups_conv2d_2_split_conv2d__)


def convert_gelu_to_approximate(model, target_activation="gelu/app"):
    from keras_cv_attention_models import common_layers

    def __convert_gelu_to_approximate__(layer):
        if isinstance(layer, layers.Activation) and layer.activation.__name__ == "gelu":
            if target_activation == "gelu/quick":
                return layers.Lambda(common_layers.gelu_quick)
            else:  # "gelu/app"
                return layers.Lambda(lambda xx: functional.gelu(xx, approximate=True))
        return layer

    input_tensors = layers.Input(model.input_shape[1:])
    return models.clone_model(model, input_tensors=input_tensors, clone_function=__convert_gelu_to_approximate__)


def convert_extract_patches_to_conv(model):
    from keras_cv_attention_models import attention_layers

    def __convert_extract_patches_to_conv__(layer):
        if isinstance(layer, attention_layers.CompatibleExtractPatches):
            aa = layer.get_config()
            aa.update({"force_conv": True})
            bb = attention_layers.CompatibleExtractPatches.from_config(aa)
            bb.build(layer.input_shape)  # No weights for this layer
            return bb
        return layer

    input_tensors = layers.Input(model.input_shape[1:])
    return models.clone_model(model, input_tensors=input_tensors, clone_function=__convert_extract_patches_to_conv__)


def convert_gelu_and_extract_patches_for_tflite(model):
    print("[Deprecated], use convert_gelu_to_approximate -> convert_extract_patches_to_conv instead")
    model = convert_gelu_to_approximate(model)
    model = convert_extract_patches_to_conv(model)
    return model


def convert_dense_to_conv(model):
    """Convert Dense layers to Conv1D or Conv2D, as TFLite not supporting Dense layers with xnnpack [???]
    >>> from keras_cv_attention_models import model_surgery

    >>> inputs = keras.layers.Input([32, 32, 3])
    >>> nn = keras.layers.Dense(12, use_bias=True)(inputs)
    >>> mm = keras.models.Model(inputs, nn, name='test_dense')
    >>>
    >>> bb = model_surgery.convert_dense_to_conv(mm)
    >>> inputs = tf.random.uniform([1, 32, 32, 3])
    >>> np.allclose(mm(inputs), bb(inputs))
    >>>
    >>> converter = tf.lite.TFLiteConverter.from_keras_model(bb)
    >>> open(mm.name + ".tflite", "wb").write(converter.convert())
    """

    def __convert_dense_to_conv__(layer):
        if isinstance(layer, layers.Dense) and (len(layer.input_shape) == 3 or len(layer.input_shape) == 4):
            target_layer = layers.Conv1D if len(layer.input_shape) == 3 else layers.Conv2D
            bb = target_layer(layer.units, kernel_size=1, use_bias=layer.use_bias, name=layer.name + "conv")
            bb(initializers.ones()([1, *layer.input_shape[1:]]))

            wws = [np.reshape(ii, jj.shape) for ii, jj in zip(layer.get_weights(), bb.get_weights())]
            bb.set_weights(wws)
            return bb
        return layer

    input_tensors = layers.Input(model.input_shape[1:])
    return models.clone_model(model, input_tensors=input_tensors, clone_function=__convert_dense_to_conv__)


def convert_to_fixed_batch_size(model, batch_size=1):
    """Fixing batch_size avoiding dynamic tensors, for supporting TFLite xnnpack. https://github.com/tensorflow/tensorflow/issues/60762"""
    input_tensors = [layers.Input(ii.shape[1:], dtype=ii.dtype, name=ii.name, batch_size=batch_size) for ii in model.inputs]
    return models.clone_model(model, input_tensors=input_tensors)


def prepare_for_tflite(model, batch_size=1):
    model = convert_to_fixed_batch_size(model) if batch_size > 0 else convert_dense_to_conv(model)
    # model = convert_groups_conv2d_2_split_conv2d(model)
    # model = convert_gelu_to_approximate(model)
    model = convert_extract_patches_to_conv(model)
    return model


def _tf_export_onnx_(model, filepath=None, fuse_conv_bn=True, batch_size=None, simplify=False, **kwargs):
    import tensorflow as tf
    import tf2onnx

    model = convert_extract_patches_to_conv(model)  # ExtractImagePatches not supported in ONNX
    model = convert_groups_conv2d_2_split_conv2d(model)  # StatefulPartitionCall not supported in ONNX
    if fuse_conv_bn:
        print("Fuse Conv + BN")
        model = convert_to_fused_conv_bn_model(model)

    spec = (tf.TensorSpec((batch_size, *model.input_shape[1:]), tf.float32, name="input"),)
    filepath = filepath or model.name + ".onnx"
    model_proto, _ = tf2onnx.convert.from_keras(model, input_signature=spec, output_path=filepath, **kwargs)
    print("Exported onnx:", filepath)

    if simplify:
        import onnx, onnxsim

        print("Running onnxsim.simplify...")
        model_proto, check = onnxsim.simplify(model_proto)
        if check:
            with open(filepath, "wb") as ff:
                ff.write(model_proto.SerializeToString())
            print("Exported simplified onnx:", filepath)
        else:
            print("[Error] failed to simplify onnx:", filepath)


def export_onnx(model, filepath=None, fuse_conv_bn=True, batch_size=None, simplify=False, **kwargs):
    """
    >>> # !pip install onnx tf2onnx onnxsim onnxruntime
    >>> from keras_cv_attention_models import volo, nat, model_surgery
    >>> mm = nat.DiNAT_Small(pretrained=True)
    >>> model_surgery.export_onnx(mm, fuse_conv_bn=True, batch_size=1, simplify=True)
    >>> # Exported simplified onnx: dinat_small.onnx

    # Run test
    >>> from keras_cv_attention_models.imagenet import eval_func
    >>> aa = eval_func.ONNXModelInterf(mm.name + '.onnx')
    >>> inputs = np.random.uniform(size=[1, *mm.input_shape[1:]]).astype('float32')
    >>> print(f"{np.allclose(aa(inputs), mm(inputs), atol=1e-5) = }")
    >>> # np.allclose(aa(inputs), mm(inputs), atol=1e-5) = True
    """
    if hasattr(model, "export_onnx"):
        model.export_onnx(filepath=filepath, simplify=simplify, **kwargs)
    else:
        _tf_export_onnx_(model, filepath=filepath, fuse_conv_bn=fuse_conv_bn, batch_size=batch_size, simplify=simplify, **kwargs)


""" Convert previous weights to new version """


def swin_convert_pos_emb_mlp_to_MlpPairwisePositionalEmbedding_weights(source_file, save_path):
    """
    Convert previous kecam <= 1.3.18 Swinv2 weights using PairWiseRelativePositionalEmbedding -> mlp_block
    to MlpPairwisePositionalEmbedding layer.

    Example:
    >>> from keras_cv_attention_models.model_surgery import swin_convert_pos_emb_mlp_to_MlpPairwisePositionalEmbedding_weights
    >>> source_file = os.path.expanduser("~/.keras/models/swin_transformer_v2_tiny_window16_256_imagenet.h5")
    >>> target_file = "swin_transformer_v2_tiny_window16_256_imagenet_modified.h5"
    >>> swin_convert_pos_emb_mlp_to_MlpPairwisePositionalEmbedding_weights(source_file, target_file)
    >>> # Saved to: swin_transformer_v2_tiny_window16_256_imagenet_modified.h5
    # Test
    >>> from keras_cv_attention_models import test_images, swin_transformer_v2
    >>> mm = swin_transformer_v2.SwinTransformerV2Tiny_window16(pretrained=target_file)
    >>> mm.decode_predictions(mm(mm.preprocess_input(test_images.cat())))
    """
    from keras_cv_attention_models import download_and_load

    with download_and_load.H5orKerasFileReader(source_file) as ff:
        bb = {kk: np.array(vv) for kk, vv in ff.items()}
    layer_names = list(bb.keys())

    pos_emb_weight_names = ["hidden_weight:0", "hidden_bias:0", "out:0"]
    for layer_name in layer_names:
        if layer_name.endswith("_meta_dense_1"):
            dense_1_name, dense_2_name, pos_layer_name = layer_name, layer_name.replace("meta_dense_1", "meta_dense_2")
            pos_layer_name = layer_name.replace("meta_dense_1", "pos_emb")
            if dense_1_name not in bb or dense_2_name not in bb:
                print("[WARNING] cannot match {}, requires source containing {} and {}".format(pos_layer_name, dense_1_name, dense_2_name))
                continue

            print("Layer:", pos_layer_name)
            dense_1 = bb.pop(dense_1_name)
            dense_2 = bb.pop(dense_2_name)
            weight_names = [pos_layer_name + "/" + ii for ii in pos_emb_weight_names]
            print("    weight_names:", weight_names, "weight shapes:", [ii.shape for ii in dense_1 + dense_2])
            bb[pos_layer_name] = dict(zip(weight_names, dense_1 + dense_2))
    download_and_load.save_weights_to_hdf5_file(save_path, bb)
    print(">>>> Saved to:", save_path)



import numpy as np
from keras_cv_attention_models import backend
from keras_cv_attention_models.backend import layers, functional, initializers, image_data_format
from keras_cv_attention_models.aotnet import AotNet
from keras_cv_attention_models.models import register_model
from keras_cv_attention_models.attention_layers import conv2d_no_bias, scaled_dot_product_attention, qkv_to_multi_head_channels_last_format
from keras_cv_attention_models.download_and_load import reload_model_weights

BATCH_NORM_DECAY = 0.9
BATCH_NORM_EPSILON = 1e-5

PRETRAINED_DICT = {
    # "botnet50": {"imagenet": "b221b45ca316166fc858fda1cf4fd946"},
    "botnet26t": {"imagenet": {256: "6d7a9548f866b4971ca2c9d17dd815fc"}},
    "botnext_eca26t": {"imagenet": {256: "170b9b4d7fba88dbcb41716047c047b9"}},
    "botnet_se33t": {"imagenet": {256: "f612743ec59d430f197bc38b3a7f8837"}},
}


@backend.register_keras_serializable(package="botnet")
class RelativePositionalEmbedding(layers.Layer):
    def __init__(self, position_height=0, position_width=0, use_absolute_pos=False, dynamic_shape=False, **kwargs):
        super().__init__(**kwargs)
        self.position_height = position_height
        self.position_width = position_width if position_width > 0 else position_height
        self.use_absolute_pos = use_absolute_pos
        self.dynamic_shape = dynamic_shape  # Deprecated

    def build(self, input_shape):
        _, num_heads, height, width, key_dim = input_shape
        self.position_height = self.position_height if self.position_height > height else height
        self.position_width = self.position_width if self.position_width > width else width
        self.key_dim = key_dim
        stddev = key_dim**-0.5
        self.num_heads, self.input_height, self.input_width = num_heads, height, width

        if self.use_absolute_pos:
            hh_shape = (key_dim, self.position_height)
            ww_shape = (key_dim, self.position_width)
        else:
            hh_shape = (key_dim, 2 * self.position_height - 1)
            ww_shape = (key_dim, 2 * self.position_width - 1)

        initializer = initializers.random_normal(stddev=stddev)
        self.pos_emb_h = self.add_weight(name="r_height", shape=hh_shape, initializer=initializer, trainable=True)
        self.pos_emb_w = self.add_weight(name="r_width", shape=ww_shape, initializer=initializer, trainable=True)

        super().build(input_shape)

    def get_config(self):
        base_config = super(RelativePositionalEmbedding, self).get_config()
        base_config.update(
            {
                "position_height": self.position_height,
                "position_width": self.position_width,
                "use_absolute_pos": self.use_absolute_pos,
                "dynamic_shape": self.dynamic_shape,
            }
        )
        return base_config

    def rel_to_abs(self, rel_pos, is_height):
        """
        Converts relative indexing to absolute.
        Input: [bs+heads, height, width, 2 * pos_dim - 1]
        Output: [bs+heads, height, width, pos_dim]
        """
        pos_dim = self.position_height if is_height else self.position_width  # Use static values
        num_blocks = self.input_height if is_height else self.input_width  # Use static values

        # pos_dim = (dim + 1) // 2
        if pos_dim == 1:
            return rel_pos
        if num_blocks == 1:
            return rel_pos[:, :, :, -pos_dim:]
        _, hh, ww, dim = rel_pos.shape  # [bs+heads, height, width, 2 * width - 1]
        full_rank_gap = pos_dim - num_blocks
        # [bs+heads, height, width * (2 * pos_dim - 1)] --> [bs+heads, height, width * (2 * pos_dim - 1) - width]
        flat_x = functional.reshape(rel_pos, [-1, hh, ww * dim])[:, :, ww - 1 : -1]
        # [bs+heads, height, width, 2 * (pos_dim - 1)] --> [bs+heads, height, width, pos_dim]
        # print(f">>>> {full_rank_gap = }, {flat_x.shape = }")
        return functional.reshape(flat_x, [-1, hh, ww, 2 * (pos_dim - 1)])[:, :, :, full_rank_gap : pos_dim + full_rank_gap]

    def relative_logits(self, inputs):
        # bs, heads, hh, ww, cc = inputs.shape  # e.g.: [1, 4, 14, 16, 128]
        inputs = functional.reshape(inputs, [-1, self.input_height, self.input_width, self.key_dim])  # Merge bs and heads, for supporting TFLite conversion
        rel_logits_w = functional.matmul(inputs, self.pos_emb_w)  # [4, 14, 16, 31], 2 * 16 - 1 == 31
        rel_logits_w = self.rel_to_abs(rel_logits_w, is_height=False)  # [4, 14, 16, 16]

        query_h = functional.transpose(inputs, [0, 2, 1, 3])  # [4, 16, 14, 128], [bs+heads, ww, hh, dims], Exchange `ww` and `hh`
        rel_logits_h = functional.matmul(query_h, self.pos_emb_h)  # [4, 16, 14, 27], 2 * 14 - 1 == 27
        rel_logits_h = self.rel_to_abs(rel_logits_h, is_height=True)  # [4, 16, 14, 14]
        rel_logits_h = functional.transpose(rel_logits_h, [0, 2, 1, 3])  # [4, 14, 16, 14], transpose back

        logits = functional.expand_dims(rel_logits_w, axis=-2) + functional.expand_dims(rel_logits_h, axis=-1)  # [4, 14, 16, 14, 16]
        return functional.reshape(logits, [-1, self.num_heads, self.input_height, self.input_width, self.position_height, self.position_width])

    def absolute_logits(self, inputs):
        # pos_emb = tf.expand_dims(self.pos_emb_w, -2) + tf.expand_dims(self.pos_emb_h, -1)
        # return tf.einsum("bxyhd,dpq->bhxypq", inputs, pos_emb)
        rel_logits_w = functional.matmul(inputs, self.pos_emb_w)
        rel_logits_h = functional.matmul(inputs, self.pos_emb_h)
        return functional.expand_dims(rel_logits_w, axis=-2) + functional.expand_dims(rel_logits_h, axis=-1)

    def call(self, inputs):
        pos_emb = self.absolute_logits(inputs) if self.use_absolute_pos else self.relative_logits(inputs)
        if self.dynamic_shape:
            _, _, hh, ww, _ = inputs.shape
            if hh < self.position_height or ww < self.position_width:
                pos_emb = pos_emb[:, :, :, :, :hh, :ww]
        return pos_emb

    def load_resized_weights(self, source_layer, method="bilinear"):
        # For input 224 --> [128, 27], convert to 480 --> [128, 30]
        if isinstance(source_layer, dict):
            source_pos_emb_h, source_pos_emb_w = list(source_layer.values())
        else:
            source_pos_emb_h, source_pos_emb_w = source_layer.pos_emb_h, source_layer.pos_emb_w  # layer
        source_pos_emb_h = np.array(source_pos_emb_h.detach() if hasattr(source_pos_emb_h, "detach") else source_pos_emb_h).astype("float32")
        source_pos_emb_w = np.array(source_pos_emb_w.detach() if hasattr(source_pos_emb_w, "detach") else source_pos_emb_w).astype("float32")

        image_like_h = np.expand_dims(np.transpose(source_pos_emb_h, [1, 0]), 0)
        resize_h = backend.numpy_image_resize(image_like_h, target_shape=(1, self.pos_emb_h.shape[1]), method=method)[0]
        resize_h = np.transpose(resize_h, [1, 0])

        image_like_w = np.expand_dims(np.transpose(source_pos_emb_w, [1, 0]), 0)
        resize_w = backend.numpy_image_resize(image_like_w, target_shape=(1, self.pos_emb_w.shape[1]), method=method)[0]
        resize_w = np.transpose(resize_w, [1, 0])

        self.set_weights([resize_h, resize_w])

    def show_pos_emb(self, base_size=4):
        import matplotlib.pyplot as plt

        pos_emb_h = self.pos_emb_h.detach() if hasattr(self.pos_emb_h, "detach") else self.pos_emb_h
        pos_emb_h = self.pos_emb_h.numpy() if hasattr(self.pos_emb_h, "numpy") else np.array(self.pos_emb_h)
        pos_emb_w = self.pos_emb_w.detach() if hasattr(self.pos_emb_w, "detach") else self.pos_emb_w
        pos_emb_w = self.pos_emb_w.numpy() if hasattr(self.pos_emb_w, "numpy") else np.array(self.pos_emb_w)

        fig, axes = plt.subplots(1, 3, figsize=(base_size * 3, base_size * 1))
        axes[0].imshow(pos_emb_h)
        axes[1].imshow(pos_emb_w)
        hh_sum = np.ones([1, pos_emb_h.shape[0]]) @ pos_emb_h
        ww_sum = np.ones([1, pos_emb_w.shape[0]]) @ pos_emb_w
        axes[2].imshow(np.transpose(hh_sum) + ww_sum)
        titles = ["pos_emb_h", "pos_emb_w", "sum"]
        for ax, title in zip(axes.flatten(), titles):
            ax.set_title(title)
            ax.set_axis_off()
        fig.tight_layout()
        return fig


def mhsa_with_relative_position_embedding(
    inputs, num_heads=4, key_dim=0, relative=True, out_shape=None, out_weight=True, out_bias=False, attn_dropout=0, data_format=None, name=None
):
    data_format = image_data_format() if data_format is None else data_format
    channel_axis = -1 if data_format == "channels_last" else 1
    input_channels = inputs.shape[channel_axis]
    hh, ww = inputs.shape[1:-1] if data_format == "channels_last" else inputs.shape[2:]

    key_dim = key_dim if key_dim > 0 else input_channels // num_heads
    out_shape = input_channels if out_shape is None or not out_weight else out_shape
    qk_out = num_heads * key_dim
    vv_dim = out_shape // num_heads

    # qkv = layers.Dense(qk_out * 2 + out_shape, use_bias=False, name=name and name + "qkv")(inputs)
    qkv = conv2d_no_bias(inputs, qk_out * 2 + out_shape, kernel_size=1, name=name and name + "qkv_")
    # qkv = functional.reshape(qkv, [-1, inputs.shape[1] * inputs.shape[2], qkv.shape[-1]])
    query, key, value = functional.split(qkv, [qk_out, qk_out, out_shape], axis=channel_axis)
    query, key, value = qkv_to_multi_head_channels_last_format(query, key, value, num_heads, data_format=data_format)

    # pos_query = [batch, num_heads, hh, ww, key_dim]
    pos_query = functional.reshape(query, [-1, num_heads, hh, ww, key_dim])
    pos_emb = RelativePositionalEmbedding(use_absolute_pos=not relative, name=name and name + "pos_emb")(pos_query)

    output_shape = [hh, ww, out_shape]
    pos_emb_func = lambda attention_scores: attention_scores + functional.reshape(pos_emb, [-1, *attention_scores.shape[1:]])
    out = scaled_dot_product_attention(query, key, value, output_shape, pos_emb_func, out_weight, out_bias, dropout=attn_dropout, name=name)
    return out if data_format == "channels_last" else layers.Permute([3, 1, 2])(out)


def BotNet(input_shape=(224, 224, 3), strides=1, pretrained="imagenet", **kwargs):
    attn_types = [None, None, None, "bot"]
    attn_params = {"num_heads": 4, "out_weight": False}
    strides = strides if isinstance(strides, (list, tuple)) else [1, 2, 2, strides]

    model = AotNet(input_shape=input_shape, attn_types=attn_types, attn_params=attn_params, strides=strides, **kwargs)
    reload_model_weights(model, PRETRAINED_DICT, "botnet", pretrained, RelativePositionalEmbedding)
    return model


@register_model
def BotNet50(input_shape=(224, 224, 3), num_classes=1000, activation="relu", classifier_activation="softmax", pretrained="imagenet", strides=1, **kwargs):
    num_blocks = [3, 4, 6, 3]
    return BotNet(**locals(), model_name="botnet50", **kwargs)


@register_model
def BotNet101(input_shape=(224, 224, 3), num_classes=1000, activation="relu", classifier_activation="softmax", pretrained=None, strides=1, **kwargs):
    num_blocks = [3, 4, 23, 3]
    return BotNet(**locals(), model_name="botnet101", **kwargs)


@register_model
def BotNet152(input_shape=(224, 224, 3), num_classes=1000, activation="relu", classifier_activation="softmax", pretrained=None, strides=1, **kwargs):
    num_blocks = [3, 8, 36, 3]
    return BotNet(**locals(), model_name="botnet152", **kwargs)


@register_model
def BotNet26T(input_shape=(256, 256, 3), num_classes=1000, activation="relu", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    num_blocks = [2, 2, 2, 2]
    attn_types = [None, None, [None, "bot"], "bot"]
    attn_params = {"num_heads": 4, "out_weight": False}
    stem_type = "tiered"

    model = AotNet(model_name="botnet26t", **locals(), **kwargs)
    reload_model_weights(model, PRETRAINED_DICT, "botnet", pretrained, RelativePositionalEmbedding)
    return model


@register_model
def BotNextECA26T(input_shape=(256, 256, 3), num_classes=1000, activation="swish", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    num_blocks = [2, 2, 2, 2]
    attn_types = [None, None, [None, "bot"], "bot"]
    attn_params = {"num_heads": 4, "key_dim": 16, "out_weight": False}
    use_eca = True
    group_size = 16
    stem_type = "tiered"
    model = AotNet(model_name="botnext_eca26t", **locals(), **kwargs)
    reload_model_weights(model, PRETRAINED_DICT, "botnet", pretrained, RelativePositionalEmbedding)
    return model


@register_model
def BotNetSE33T(input_shape=(256, 256, 3), num_classes=1000, activation="swish", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    stem_type = "tiered"
    stem_last_strides = 2
    stem_downsample = False
    out_channels = [256, 512, 1024, 1536]
    hidden_channel_ratio = [1 / 4, 1 / 4, 1 / 4, 1 / 3]
    num_blocks = [2, 3, 3, 2]
    attn_types = [None, [None, None, "bot"], [None, None, "bot"], "bot"]
    attn_params = {"num_heads": 4, "out_weight": False}
    se_ratio = 1 / 16
    output_num_features = 1280

    model = AotNet(model_name="botnet_se33t", **locals(), **kwargs)
    reload_model_weights(model, PRETRAINED_DICT, "botnet", pretrained, RelativePositionalEmbedding)
    return model



from keras_cv_attention_models.beit.beit import Beit, keras_model_load_weights_from_pytorch_model
from keras_cv_attention_models.models import register_model


def ViT(attn_qv_bias=False, attn_qkv_bias=True, use_abs_pos_emb=True, layer_scale=0, use_mean_pooling_head=False, model_name="vit", **kwargs):
    kwargs.pop("kwargs", None)
    return Beit(**locals(), **kwargs)


def ViTText(
    vocab_size=49408,
    max_block_size=77,
    text_positional_dropout=0,
    text_use_positional_embedding=True,
    include_top=True,
    layer_norm_epsilon=1e-5,
    activation="gelu/quick",
    model_name="vit_text",
    **kwargs,
):
    attn_qv_bias = kwargs.pop("attn_qv_bias", False)
    attn_qkv_bias = kwargs.pop("attn_qkv_bias", True)
    use_abs_pos_emb = kwargs.pop("use_abs_pos_emb", True)
    layer_scale = kwargs.pop("layer_scale", 0)
    use_mean_pooling_head = kwargs.pop("use_mean_pooling_head", False)
    kwargs.pop("kwargs", None)
    return Beit(**locals(), **kwargs)


@register_model
def ViTTinyPatch16(input_shape=(196, 196, 3), num_classes=1000, activation="gelu", classifier_activation="softmax", pretrained="imagenet21k-ft1k", **kwargs):
    embed_dim = 192
    depth = 12
    num_heads = 3
    patch_size = kwargs.pop("patch_size", 16)
    return ViT(**locals(), model_name="vit_tiny_patch16", **kwargs)


@register_model
def ViTBasePatch16(input_shape=(196, 196, 3), num_classes=1000, activation="gelu", classifier_activation="softmax", pretrained="imagenet21k-ft1k", **kwargs):
    embed_dim = 768
    depth = 12
    num_heads = 12
    patch_size = kwargs.pop("patch_size", 16)
    return ViT(**locals(), model_name="vit_base_patch16", **kwargs)


@register_model
def ViTLargePatch14(input_shape=(196, 196, 3), num_classes=1000, activation="gelu", classifier_activation="softmax", pretrained="imagenet21k-ft1k", **kwargs):
    embed_dim = 1024
    depth = 24
    num_heads = 16
    patch_size = kwargs.pop("patch_size", 14)
    return ViT(**locals(), model_name="vit_large_patch14", **kwargs)


@register_model
def ViTTextLargePatch14(vocab_size=49408, max_block_size=77, activation="gelu/quick", include_top=True, pretrained="clip", **kwargs):
    embed_dim = 768
    depth = 12
    num_heads = 12
    patch_size = kwargs.pop("patch_size", 14)
    return ViTText(**locals(), model_name="vit_text_large_patch14", **kwargs)



from keras_cv_attention_models import backend
from keras_cv_attention_models.backend import layers, functional, models, image_data_format
from keras_cv_attention_models.models import register_model
from keras_cv_attention_models.attention_layers import (
    activation_by_name,
    batchnorm_with_activation,
    conv2d_no_bias,
    depthwise_conv2d_no_bias,
    drop_block,
    se_module,
    window_attention,
    MultiHeadPositionalEmbedding,
    add_pre_post_process,
)
from keras_cv_attention_models.download_and_load import reload_model_weights

PRETRAINED_DICT = {
    "efficientvit_m0": {"imagenet": {224: "3a3855160a3262ddcb7969b8fc8c1dcc"}},
    "efficientvit_m1": {"imagenet": {224: "ea8f5812ea397b821abfc83cddf5845b"}},
    "efficientvit_m2": {"imagenet": {224: "6d267ec3419b7257f293273cef27dfc8"}},
    "efficientvit_m3": {"imagenet": {224: "6f3e5903fb230a83c2ca17761357f15f"}},
    "efficientvit_m4": {"imagenet": {224: "ada36562f1702b3baa3578e9d9a1cf7e"}},
    "efficientvit_m5": {"imagenet": {224: "7277f60414af8111f86d19d486a7f301"}},
}


def cascaded_mhsa_with_multi_head_position(inputs, num_heads=4, key_dim=-1, kernel_sizes=5, activation="relu", name=None):
    height, width = inputs.shape[1:-1] if image_data_format() == "channels_last" else inputs.shape[2:]
    channel_axis = -1 if image_data_format() == "channels_last" else 1
    key_dim = key_dim if key_dim > 0 else inputs.shape[channel_axis] // num_heads
    out_shape = inputs.shape[channel_axis]
    value_dim = out_shape // num_heads
    qk_scale = 1.0 / (float(key_dim) ** 0.5)
    # embed_dim = key_dim * num_heads

    qkv_dim = key_dim + key_dim + value_dim
    kernel_sizes = kernel_sizes if isinstance(kernel_sizes, (list, tuple)) else [kernel_sizes] * num_heads
    head_inputs = functional.split(inputs, num_heads, axis=channel_axis)
    output = []
    for id, (head_input, kernel_size) in enumerate(zip(head_inputs, kernel_sizes)):
        head_input = head_input if id == 0 else (head_input + output[-1])
        cur_name = name and name + "{}_".format(id + 1)
        qkv = conv2d_no_bias(head_input, qkv_dim, name=cur_name and cur_name + "qkv_")
        qkv = batchnorm_with_activation(qkv, activation=None, name=cur_name and cur_name + "qkv_")
        qq, kk, vv = functional.split(qkv, [key_dim, key_dim, value_dim], axis=channel_axis)

        qq = depthwise_conv2d_no_bias(qq, kernel_size=kernel_size, padding="same", name=cur_name and cur_name + "query_")
        qq = batchnorm_with_activation(qq, activation=None, name=cur_name and cur_name + "query_")

        if image_data_format() == "channels_last":
            qq = functional.reshape(qq, [-1, qq.shape[1] * qq.shape[2], qq.shape[-1]])
            kk = functional.transpose(functional.reshape(kk, [-1, kk.shape[1] * kk.shape[2], kk.shape[-1]]), [0, 2, 1])
            vv = functional.reshape(vv, [-1, vv.shape[1] * vv.shape[2], vv.shape[-1]])
        else:
            qq = functional.transpose(functional.reshape(qq, [-1, qq.shape[1], qq.shape[2] * qq.shape[3]]), [0, 2, 1])
            kk = functional.reshape(kk, [-1, kk.shape[1], kk.shape[2] * kk.shape[3]])
            vv = functional.reshape(vv, [-1, vv.shape[1], vv.shape[2] * vv.shape[3]])

        # attention_scores = layers.Lambda(lambda xx: functional.matmul(xx[0], xx[1]))([query, key]) * qk_scale  # [batch, num_heads, key_dim, key_dim]
        attention_scores = (qq @ kk) * qk_scale
        # print(f"{height = }, {width = }")
        attention_scores = MultiHeadPositionalEmbedding(query_height=height, name=cur_name and cur_name + "attn_pos")(attention_scores)
        attention_scores = layers.Softmax(axis=-1, name=cur_name and cur_name + "attention_scores")(attention_scores)
        # attention_scores = layers.Dropout(attn_dropout, name=name and name + "attn_drop")(attention_scores) if attn_dropout > 0 else attention_scores

        if image_data_format() == "channels_last":
            attention_output = attention_scores @ vv
            head_output = functional.reshape(attention_output, [-1, height, width, value_dim])
        else:
            attention_output = vv @ functional.transpose(attention_scores, [0, 2, 1])
            head_output = functional.reshape(attention_output, [-1, value_dim, height, width])
        output.append(head_output)
    output = functional.concat(output, axis=channel_axis)

    output = activation_by_name(output, activation=activation, name=name)
    output = conv2d_no_bias(output, out_shape, name=name and name + "out")
    output = batchnorm_with_activation(output, activation=None, zero_gamma=True, name=name and name + "out_")
    return output


def res_depthwise_ffn(inputs, mlp_ratio=2, drop_rate=0, activation="relu", name=""):
    input_channel = inputs.shape[-1 if image_data_format() == "channels_last" else 1]
    dw = depthwise_conv2d_no_bias(inputs, kernel_size=3, strides=1, padding="same", name=name + "dw_")
    dw = batchnorm_with_activation(dw, activation=None, zero_gamma=True, name=name + "dw_")
    dw = inputs + drop_block(dw, drop_rate)

    ffn = conv2d_no_bias(dw, input_channel * mlp_ratio, name=name + "ffn_1_")
    ffn = batchnorm_with_activation(ffn, activation=activation, zero_gamma=True, name=name + "ffn_1_")
    ffn = conv2d_no_bias(ffn, input_channel, name=name + "ffn_2_")
    ffn = batchnorm_with_activation(ffn, activation=None, zero_gamma=True, name=name + "ffn_2_")
    ffn = drop_block(ffn, drop_rate)
    return layers.Add(name=name + "output")([dw, ffn])


def attn_block(inputs, window_size=7, num_heads=4, key_dim=16, kernel_sizes=5, mlp_ratio=2, drop_rate=0, activation="relu", name=""):
    height, width = inputs.shape[1:-1] if image_data_format() == "channels_last" else inputs.shape[2:]
    pre = res_depthwise_ffn(inputs, mlp_ratio=mlp_ratio, drop_rate=drop_rate, activation=activation, name=name + "pre_")

    attn_kw = {"num_heads": num_heads, "key_dim": key_dim, "kernel_sizes": kernel_sizes, "activation": activation, "name": name + "attn_"}
    if height <= window_size and width <= window_size:
        attn = cascaded_mhsa_with_multi_head_position(pre, **attn_kw)
    else:
        data_format = image_data_format()
        attn = window_attention(pre, window_size=window_size, attention_block=cascaded_mhsa_with_multi_head_position, data_format=data_format, **attn_kw)
    attn = pre + drop_block(attn, drop_rate)

    return res_depthwise_ffn(attn, mlp_ratio=mlp_ratio, drop_rate=drop_rate, activation=activation, name=name + "post_")


def down_sample_block(inputs, out_channel, hiddem_raio=4, activation="relu", name=""):
    input_channel = inputs.shape[-1 if image_data_format() == "channels_last" else 1]
    nn = conv2d_no_bias(inputs, input_channel * hiddem_raio, kernel_size=1, strides=1, padding="same", name=name)
    nn = batchnorm_with_activation(nn, activation=activation, name=name + "pw_")
    nn = depthwise_conv2d_no_bias(nn, kernel_size=3, strides=2, padding="same", name=name)
    nn = batchnorm_with_activation(nn, activation=activation, name=name + "dw_")
    nn = se_module(nn, activation=activation, name=name + "se_")
    nn = conv2d_no_bias(nn, out_channel, kernel_size=1, strides=1, padding="same", name=name + "out_")
    nn = batchnorm_with_activation(nn, activation=None, name=name + "out_")
    return nn


def EfficientViT_M(
    num_blocks=[1, 2, 3],
    out_channels=[64, 128, 192],
    window_size=7,
    num_heads=4,
    key_dim=16,
    kernel_sizes=5,
    mlp_ratio=2,
    stem_width=-1,
    input_shape=(224, 224, 3),
    num_classes=1000,
    activation="relu",
    drop_connect_rate=0,
    classifier_activation="softmax",
    use_distillation=False,  # not provided in pretrained weights
    dropout=0,
    pretrained=None,
    model_name="efficientvit_m",
    kwargs=None,
):
    # Regard input_shape as force using original shape if len(input_shape) == 4,
    # else assume channel dimension is the one with min value in input_shape, and put it first or last regarding image_data_format
    input_shape = backend.align_input_shape_by_image_data_format(input_shape)
    inputs = layers.Input(input_shape)
    stem_width = stem_width if stem_width > 0 else out_channels[0]
    nn = conv2d_no_bias(inputs, stem_width // 8, 3, strides=2, padding="same", name="stem_1_")
    nn = batchnorm_with_activation(nn, activation=activation, name="stem_1_")
    nn = conv2d_no_bias(nn, stem_width // 4, 3, strides=2, padding="same", name="stem_2_")
    nn = batchnorm_with_activation(nn, activation=activation, name="stem_2_")
    nn = conv2d_no_bias(nn, stem_width // 2, 3, strides=2, padding="same", name="stem_3_")
    nn = batchnorm_with_activation(nn, activation=activation, name="stem_3_")
    nn = conv2d_no_bias(nn, stem_width, 3, strides=2, padding="same", name="stem_4_")
    nn = batchnorm_with_activation(nn, activation=None, name="stem_4_")

    """ stages """
    total_blocks = sum(num_blocks)
    global_block_id = 0
    for stack_id, (num_block, out_channel) in enumerate(zip(num_blocks, out_channels)):
        stack_name = "stack{}_".format(stack_id + 1)
        if stack_id > 0:
            nn = res_depthwise_ffn(nn, mlp_ratio=mlp_ratio, activation=activation, name=stack_name + "pre_")
            nn = down_sample_block(nn, out_channel, activation=activation, name=stack_name + "downsample_")
            nn = res_depthwise_ffn(nn, mlp_ratio=mlp_ratio, activation=activation, name=stack_name + "post_")

        cur_num_heads = num_heads[stack_id] if isinstance(num_heads, (list, tuple)) else num_heads
        for block_id in range(num_block):
            block_name = stack_name + "block{}_".format(block_id + 1)
            block_drop_rate = drop_connect_rate * global_block_id / total_blocks
            nn = attn_block(nn, window_size, cur_num_heads, key_dim, kernel_sizes, mlp_ratio, drop_rate=block_drop_rate, activation=activation, name=block_name)
            global_block_id += 1

    """ output """
    if num_classes > 0:
        nn = layers.GlobalAveragePooling2D(keepdims=True)(nn)  # tf.reduce_mean(nn, axis=1)
        if dropout > 0 and dropout < 1:
            nn = layers.Dropout(dropout)(nn)
        out = batchnorm_with_activation(nn, activation=None, name="pre_output_")
        out = layers.Flatten()(out)
        out = layers.Dense(num_classes, dtype="float32", activation=classifier_activation, name="head")(out)

        if use_distillation:
            distill_out = batchnorm_with_activation(nn, activation=None, name="distill_output_")
            distill_out = layers.Flatten()(distill_out)
            distill_out = layers.Dense(num_classes, dtype="float32", activation=classifier_activation, name="distill_head")(distill_out)
            out = [out, distill_out]
    else:
        out = nn

    model = models.Model(inputs, out, name=model_name)
    max_shape = max(input_shape[:2] if image_data_format() == "channels_last" else input_shape[1:])  # Relaod mismatch only if all shape < 7 * 32 = 224
    reload_model_weights(model, PRETRAINED_DICT, "efficientvit", pretrained, MultiHeadPositionalEmbedding if window_size != 7 or max_shape < 224 else None)

    add_pre_post_process(model, rescale_mode="torch")
    model.switch_to_deploy = lambda: switch_to_deploy(model)
    return model


def switch_to_deploy(model):
    from keras_cv_attention_models.model_surgery.model_surgery import fuse_distill_head

    new_model = fuse_distill_head(model, head_bn="pre_output_bn", distill_head_bn="distill_output_bn") if "head" in model.output_names else model
    add_pre_post_process(new_model, rescale_mode=model.preprocess_input.rescale_mode, post_process=model.decode_predictions)
    return new_model


@register_model
def EfficientViT_M0(input_shape=(224, 224, 3), num_classes=1000, activation="relu", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    return EfficientViT_M(**locals(), model_name="efficientvit_m0", **kwargs)


@register_model
def EfficientViT_M1(input_shape=(224, 224, 3), num_classes=1000, activation="relu", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    out_channels = [128, 144, 192]
    num_heads = [2, 3, 3]
    kernel_sizes = [7, 5, 3, 3]
    return EfficientViT_M(**locals(), model_name="efficientvit_m1", **kwargs)


@register_model
def EfficientViT_M2(input_shape=(224, 224, 3), num_classes=1000, activation="relu", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    out_channels = [128, 192, 224]
    num_heads = [4, 3, 2]
    kernel_sizes = [7, 5, 3, 3]
    return EfficientViT_M(**locals(), model_name="efficientvit_m2", **kwargs)


@register_model
def EfficientViT_M3(input_shape=(224, 224, 3), num_classes=1000, activation="relu", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    out_channels = [128, 240, 320]
    num_heads = [4, 3, 4]
    kernel_sizes = 5
    return EfficientViT_M(**locals(), model_name="efficientvit_m3", **kwargs)


@register_model
def EfficientViT_M4(input_shape=(224, 224, 3), num_classes=1000, activation="relu", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    out_channels = [128, 256, 384]
    num_heads = 4
    kernel_sizes = [7, 5, 3, 3]
    return EfficientViT_M(**locals(), model_name="efficientvit_m4", **kwargs)


@register_model
def EfficientViT_M5(input_shape=(224, 224, 3), num_classes=1000, activation="relu", classifier_activation="softmax", pretrained="imagenet", **kwargs):
    out_channels = [192, 288, 384]
    num_blocks = [1, 3, 4]
    num_heads = [3, 3, 4]
    kernel_sizes = [7, 5, 3, 3]
    return EfficientViT_M(**locals(), model_name="efficientvit_m5", **kwargs)



import numpy as np


def get_plot_cols_rows(total, rows=-1, ceil_mode=False):
    if rows == -1 and total < 8:
        rows = 1  # for total in [1, 7], plot 1 row only
    elif rows == -1:
        rr = int(np.floor(np.sqrt(total)))
        for ii in range(1, rr + 1)[::-1]:
            if total % ii == 0:
                rows = ii
                break
    if ceil_mode:
        cols = int(np.ceil(total / rows))
    else:
        cols = total // rows
    return cols, rows


def put_text_on_image(image, text, coord=(5, 5), color=(255, 0, 0)):
    from PIL import Image
    from PIL import ImageDraw

    # from PIL import ImageFont

    image = image * 255 if image.max() < 2 else image
    img = Image.fromarray(image.astype("uint8"))
    draw = ImageDraw.Draw(img)
    draw.text(coord, str(text), color)
    return np.array(img)


def stack_and_plot_images(images, texts=None, margin=5, margin_value=0, rows=-1, ax=None, base_size=3):
    """Stack and plot a list of images. Returns ax, stacked_images"""
    import matplotlib.pyplot as plt

    cols, rows = get_plot_cols_rows(len(images), rows, ceil_mode=True)
    if cols * rows > len(images):
        padded = cols * rows - len(images)
        images += [np.zeros_like(images[-1]) + 255 for ii in range(padded)]
    # images = images[: rows * cols]
    # print(">>>> rows:", rows, ", cols:", cols, ", total:", len(images))

    if texts is not None:
        images = [put_text_on_image(imm, itt) for imm, itt in zip(images, texts)] + list(images[len(texts) :])
        images = np.array(images)

    if ax is None:
        _, ax = plt.subplots(1, 1, figsize=(base_size * cols, base_size * rows))

    if margin > 0:
        ww_margin = np.zeros_like(images[0][:, :margin]) + margin_value
        ww_margined_images = [np.hstack([ii, ww_margin]) for ii in images]
        hstacked_images = [np.hstack(ww_margined_images[ii : ii + cols]) for ii in range(0, len(ww_margined_images), cols)]

        hh_margin = np.zeros_like(hstacked_images[0][:margin]) + margin_value
        hh_margined_images = [np.vstack([ii, hh_margin]) for ii in hstacked_images]
        vstacked_images = np.vstack(hh_margined_images)

        stacked_images = vstacked_images[:-margin, :-margin]
    else:
        stacked_images = np.vstack([np.hstack(images[ii * cols : (ii + 1) * cols]) for ii in range(rows)])

    ax.imshow(stacked_images)
    ax.set_axis_off()
    ax.grid(False)
    plt.tight_layout()
    # plt.show()
    return ax, stacked_images


""" Show detection results """


def draw_bboxes(bboxes, ax=None):
    import matplotlib.pyplot as plt
    import numpy as np

    if ax is None:
        fig, ax = plt.subplots()
    bboxes = np.array(bboxes).astype("int32")
    for bb in bboxes:
        ax.plot(bb[[1, 1, 3, 3, 1]], bb[[0, 2, 2, 0, 0]])
    plt.show()
    return ax


def show_image_with_bboxes(
    image, bboxes, labels=None, confidences=None, is_bbox_width_first=False, ax=None, label_font_size=8, num_classes=80, indices_2_labels=None
):
    import matplotlib.pyplot as plt
    import numpy as np
    from keras_cv_attention_models.coco import info

    need_plt_show = False
    if ax is None:
        fig, ax = plt.subplots()
        need_plt_show = True
    ax.imshow(image)
    bboxes = np.array(bboxes)
    if is_bbox_width_first:
        bboxes = bboxes[:, [1, 0, 3, 2]]
    for id, bb in enumerate(bboxes):
        # bbox is [top, left, bottom, right]
        bb = [bb[0] * image.shape[0], bb[1] * image.shape[1], bb[2] * image.shape[0], bb[3] * image.shape[1]]
        bb = np.array(bb).astype("int32")
        ax.plot(bb[[1, 1, 3, 3, 1]], bb[[0, 2, 2, 0, 0]])

        if labels is not None:
            label = int(labels[id])
            if indices_2_labels is not None:
                label = indices_2_labels.get(label, indices_2_labels.get(str(label), "None"))
            elif num_classes == 90:
                label = info.COCO_90_LABEL_DICT[label]
            elif num_classes == 80:
                label = info.COCO_80_LABEL_DICT[label]

            if confidences is not None:
                label += ": {:.4f}".format(float(confidences[id]))
            color = ax.lines[-1].get_color()
            # ax.text(bb[1], bb[0] - 5, "label: {}, {}".format(label, info.COCO_80_LABEL_DICT[label]), color=color, fontsize=8)
            ax.text(bb[1], bb[0] - 5, label, color=color, fontsize=label_font_size)
    ax.set_axis_off()
    plt.tight_layout()
    if need_plt_show:
        plt.show()
    return ax


""" Show clip results """


def show_images_texts_similarity(images, texts, similarity, ax=None, base_size=8, title=None):
    """
    Copied and modified from: https://github.com/mlfoundations/open_clip/blob/main/docs/Interacting_with_open_clip.ipynb

    Args:
      similarity: in shape `[images.shape[0], texts.shape[0]]`

    Examples:
    >>> from keras_cv_attention_models import test_images, plot_func
    >>> images = [test_images.dog(), test_images.cat(), test_images.dog_cat()] * 3
    >>> texts = ["dog", "cat", "dog_cat"]
    >>> similarity = np.random.uniform(size=[9, 3])
    >>> _ = plot_func.show_images_texts_similarity(images, texts, similarity)
    """
    import matplotlib.pyplot as plt

    if ax is None:
        _, ax = plt.subplots(1, 1, figsize=(base_size * 1.4, base_size))  # Wider one
    font_size = 9 * base_size / 8
    yticks_size = 10 * base_size / 8

    similarity = similarity.detach() if hasattr(similarity, "detach") else similarity
    similarity = similarity.numpy() if hasattr(similarity, "numpy") else similarity
    num_images, num_texts = similarity.shape[0], similarity.shape[1]

    ax.imshow(similarity.T, vmin=0.1, vmax=0.3)
    ax.set_yticks(range(num_texts), texts, fontsize=yticks_size)
    ax.set_xticks([])
    for ii, image in enumerate(images):
        ax.imshow(image, extent=(ii - 0.5, ii + 0.5, -1.6, -0.6), origin="lower")
    for xx in range(num_images):
        for yy in range(num_texts):
            ax.text(xx, yy, f"{similarity[xx, yy]:.2f}", ha="center", va="center", size=font_size)

    for side in ["left", "top", "right", "bottom"]:
        plt.gca().spines[side].set_visible(False)

    ax.set_xlim([-0.5, num_images - 0.5])
    ax.set_ylim([num_texts + 0.5, -2])
    ax.grid(False)

    if title:
        ax.set_title(title)
    plt.tight_layout()
    # plt.show()
    return ax


""" Plot training hists """


def curve_fit(source, target_len=10, skip=5, use_recent=40):
    from scipy.optimize import curve_fit

    def func_curv(x, a, b, c, d):
        pp = np.log(x)
        # pp = 1 / x
        return a * pp**3 + b * pp**2 + c * pp + d

    recent_source = source[skip:]
    use_recent = len(source) if use_recent == -1 else use_recent
    if len(recent_source) > use_recent:
        recent_source = recent_source[-use_recent:]
    start_pos = len(source) - len(recent_source)
    popt, pcov = curve_fit(func_curv, np.arange(start_pos, len(source)), recent_source)
    return list(source[: -len(recent_source)]) + func_curv(np.arange(start_pos, len(source) + target_len), *popt).tolist()


def plot_and_peak_scatter(ax, source_array, peak_method, label, skip_first=0, color=None, va="bottom", pred_curve=0, **kwargs):
    array = source_array[skip_first:]
    for id, ii in enumerate(array):
        if np.isnan(ii):
            array[id] = array[id - 1]
    ax.plot(range(skip_first, skip_first + len(array)), array, label=label, color=color, **kwargs)
    color = ax.lines[-1].get_color() if color is None else color
    pp = peak_method(array)
    vv = array[pp]
    ax.scatter(pp + skip_first, vv, color=color, marker="v")
    # ax.text(pp + skip_first, vv, "{:.4f}".format(vv), va=va, ha="right", color=color, fontsize=fontsize, rotation=0)
    ax.text(pp + skip_first, vv, "{:.4f}".format(vv), va=va, ha="right", color=color, rotation=0)

    if pred_curve > 0:
        kwargs.pop("linestyle", None)
        pred_array = curve_fit(source_array, pred_curve)[skip_first:]
        ax.plot(range(skip_first, skip_first + len(pred_array)), pred_array, color=color, linestyle=":", **kwargs)


def plot_hists(hists, names=None, base_size=6, addition_plots=["lr"], text_va=["bottom"], skip_first=0, pred_curve=0):
    import os
    import json
    import matplotlib as mpl
    import matplotlib.pyplot as plt

    mpl.rcParams["lines.linewidth"] = base_size / 8 * 1.5  # 8 -> 1.5
    mpl.rcParams["font.size"] = base_size * 2 - 4  # 8 -> 12, 6 -> 8
    mpl.rcParams["legend.fontsize"] = base_size * 2 - 4
    mpl.rcParams["xtick.labelsize"] = base_size * 2 - 4
    mpl.rcParams["ytick.labelsize"] = base_size * 2 - 4
    mpl.rcParams["xtick.major.pad"] = base_size / 2 - 2  # 8 -> 4, 6 -> 3
    mpl.rcParams["ytick.major.pad"] = base_size / 2 - 2

    num_axes = (2 + len(addition_plots)) if addition_plots is not None and len(addition_plots) != 0 else 2
    fig, axes = plt.subplots(1, num_axes, figsize=(num_axes * base_size, base_size))
    hists = [hists] if isinstance(hists, (str, dict)) else hists
    names = names if isinstance(names, (list, tuple)) else [names]
    for id, hist in enumerate(hists):
        name = names[min(id, len(names) - 1)] if names != None else None
        cur_va = text_va[id % len(text_va)]
        if isinstance(hist, str):
            name = name if name != None else os.path.splitext(os.path.basename(hist))[0]
            with open(hist, "r") as ff:
                hist = json.load(ff)
        name = name if name != None else str(id)

        acc_key = "acc"
        if acc_key not in hist:
            all_acc_key = [ii for ii in hist.keys() if "acc" in ii and "val" not in ii]
            acc_key = "acc" if len(all_acc_key) == 0 else all_acc_key[0]

        val_acc_key = "val_acc"
        if "val_ap_ar" in hist:  # It's coco hist
            hist["val_ap_0.50:0.95"] = [ii[0] for ii in hist["val_ap_ar"]]
            val_acc_key = "val_ap 0.50:0.95"
        elif val_acc_key not in hist:
            all_val_acc_key = [ii for ii in hist.keys() if "acc" in ii and "val" in ii]
            val_acc_key = "val_acc" if len(all_val_acc_key) == 0 else all_val_acc_key[0]

        cur_pred_curve = pred_curve[min(id, len(pred_curve) - 1)] if isinstance(pred_curve, (list, tuple)) else pred_curve
        plot_and_peak_scatter(axes[0], hist["loss"], np.argmin, name + " loss", skip_first, None, cur_va, pred_curve=cur_pred_curve)
        color = axes[0].lines[-1].get_color()
        val_loss = hist.get("val_loss", [])
        if len(val_loss) > 0 and "val_loss" not in addition_plots:
            plot_and_peak_scatter(axes[0], val_loss, np.argmin, name + " val_loss", skip_first, color, cur_va, cur_pred_curve, linestyle="--")

        acc = hist.get(acc_key, [])
        if len(acc) > 0:
            plot_and_peak_scatter(axes[1], acc, np.argmax, name + " " + acc_key, skip_first, color, cur_va, cur_pred_curve)

        val_acc = hist.get(val_acc_key, [])
        if len(val_acc) > 0:
            plot_and_peak_scatter(axes[1], val_acc, np.argmax, name + " " + val_acc_key, skip_first, color, cur_va, cur_pred_curve, linestyle="--")

        if addition_plots is not None and len(addition_plots) != 0:
            for id, ii in enumerate(addition_plots):
                if len(hist.get(ii, [])) > 0:
                    peak_method = np.argmin if "loss" in ii else np.argmax
                    plot_and_peak_scatter(axes[2 + id], hist[ii], peak_method, name + " " + ii, skip_first, color, cur_va, cur_pred_curve)
    for ax in axes:
        ax.legend()
        ax.grid(True)
    fig.tight_layout()
    fig.tight_layout()  # again
    plt.show()
    return fig


""" tensorboard_parallel_coordinates_plot """


def tensorboard_parallel_coordinates_plot(dataframe, metrics_name, metrics_display_name=None, skip_columns=[], log_dir="logs/hparam_tuning"):
    """
    Simmilar results with: [Visualize the results in TensorBoard's HParams plugin](https://www.tensorflow.org/tensorboard/hyperparameter_tuning_with_hparams#4_visualize_the_results_in_tensorboards_hparams_plugin).

    Wrapped function just plotting ignoring training in the tutorial.
    The logic is using `metrics_name` specified key as metrics, using other columns as `HParams`.
    For any other detail, refer original tutorial.

    Examples:
    >>> import pandas as pd
    >>> aotnet50_imagnet_results = {
    >>>     "optimizer": ["lamb", "lamb", "adamw", "adamw", "adamw"],
    >>>     "rescale_mode": ["torch", "tf", "torch", "torch", "torch"],
    >>>     "lr_base": [8e-3, 8e-3, 4e-3, 4e-3, 8e-3],
    >>>     "weight_decay": [0.05, 0.05, 0.05, 0.02, 0.02],
    >>>     "accuracy": [78.48, 78.31, 77.92, 78.06, 78.27],
    >>> }
    >>> aa = pd.DataFrame(aotnet50_imagnet_results)

    >>> from keras_cv_attention_models import plot_func
    >>> plot_func.tensorboard_parallel_coordinates_plot(aa, 'accuracy', log_dir="logs/aotnet50_imagnet_results")
    >>> # >>>> Start tensorboard by: ! tensorboard --logdir logs/aotnet50_imagnet_results
    >>> # >>>> Then select `HPARAMS` -> `PARALLEL COORDINATES VIEW`

    >>> ! tensorboard --logdir logs/aotnet50_imagnet_results
    """
    import os
    import pandas as pd
    import tensorflow as tf
    from tensorboard.plugins.hparams import api as hp

    skip_columns = skip_columns + [metrics_name]
    to_hp_discrete = lambda column: hp.HParam(column, hp.Discrete(np.unique(dataframe[column].values).tolist()))
    hp_params_dict = {column: to_hp_discrete(column) for column in dataframe.columns if column not in skip_columns}

    if dataframe[metrics_name].values.dtype == "object":  # Not numeric
        import json

        metrics_map = {ii: id for id, ii in enumerate(np.unique(dataframe[metrics_name]))}
        description = json.dumps(metrics_map)
    else:
        metrics_map, description = None, None

    METRICS = metrics_name if metrics_display_name is None else metrics_display_name
    with tf.summary.create_file_writer(log_dir).as_default():
        metrics = [hp.Metric(METRICS, display_name=METRICS, description=description)]
        hp.hparams_config(hparams=list(hp_params_dict.values()), metrics=metrics)

    for id in dataframe.index:
        log = dataframe.iloc[id]
        hparams = {hp_unit: log[column] for column, hp_unit in hp_params_dict.items()}
        print({hp_unit.name: hparams[hp_unit] for hp_unit in hparams})
        run_dir = os.path.join(log_dir, "run-%d" % id)
        with tf.summary.create_file_writer(run_dir).as_default():
            hp.hparams(hparams)  # record the values used in this trial
            metric_item = log[metrics_name] if metrics_map is None else metrics_map[log[metrics_name]]
            tf.summary.scalar(METRICS, metric_item, step=1)

    print()
    if metrics_map is not None:
        print(">>>> metrics_map:", metrics_map)
    print(">>>> Start tensorboard by: ! tensorboard --logdir {}".format(log_dir))
    print(">>>> Then select `HPARAMS` -> `PARALLEL COORDINATES VIEW`")


""" Plot model summary """


def plot_model_summary(
    plot_series, x_label="inference_qps", y_label="acc_metrics", model_table="model_summary.csv", allow_extras=None, log_scale_x=False, ax=None
):
    """
    Args:
      plot_series: list value for filtering itmes by model_table "series" column.
      x_label: string value from column names in `model_table`, x axis values.
      y_label: string value from column names in `model_table`, y axis values.
      model_table: a csv file path or loaded pandas DataFrame.
          If DataFrame, columns ["series", "model"] and [x_label, y_label] are required.
          "series" means which model series this model belongs to, and "model" is the actual model name.
      allow_extras: list value for allowing plotting data with extra pretrained.
          Default None for plotting imagenet without any extra pretrained only.
          Special string value "all" for allowing all extras.
      log_scale_x: boolean value if setting x scale in log distribution.
      ax: plotting on specific matplotlib ax. Default None for creating a new figure.

    Examples:
    >>> from keras_cv_attention_models import plot_func
    >>> plot_series = ["convnextv2", "efficientnetv2", "efficientvit_b", "fasternet", "fastervit"]
    >>> plot_func.plot_model_summary(plot_series, x_label='inference_qps', model_table="model_summary.csv", allow_extras=None)
    >>> plt.savefig('foo.png', dpi=150)  # Save if needed, with dpi speficified

    # Using custom DataFrame
    >>> from keras_cv_attention_models import plot_func
    >>> dd = pd.DataFrame({
    >>>     "series": ['Res', 'Res', 'Res', 'Res', 'Trans', 'Trans', 'Trans', 'cc'],
    >>>     'model': ['res50', 'res101', 'res201', 'aa4', 'trans_tiny', 'trans_small', 'trans_big', 'cc1'],
    >>>     'test_key': [0.2, 0.3, 0.5, 0.55, 0.1, 0.3, 0.4, 0.2],
    >>>     'valuable': [0.8, 0.92, 0.93, 0.96, 0.71, 0.74, 0.85, 0.98],
    >>> })
    >>> plot_func.plot_model_summary(plot_series=['Res', 'trans'], x_label='test_key', y_label='valuable', model_table=dd)
    """
    import matplotlib
    import matplotlib.pyplot as plt

    # ('o', 'v', '^', '<', '>', '8', 's', 'p', '*', 'h', 'H', 'D', 'd', 'P', 'X'), excluded ["1", "2", "3", "4", "+", "x", '.', ',', '|', '_']
    markers = matplotlib.markers.MarkerStyle.filled_markers
    fontsize = 9
    pre_x_labelsize, pre_y_labelsize = matplotlib.rcParams["xtick.labelsize"], matplotlib.rcParams["ytick.labelsize"]
    matplotlib.rcParams["xtick.labelsize"], matplotlib.rcParams["ytick.labelsize"] = fontsize, fontsize

    if isinstance(model_table, str):
        import pandas as pd

        dd = pd.read_csv(model_table)
    else:
        dd = model_table
    # dd = dd[dd.category == "Recognition"]
    dd = dd[dd[x_label].notnull()]
    dd = dd[dd[y_label].notnull()]

    if ax is None:
        fig, ax = plt.subplots()

    has_extra = "extra" in dd.columns
    plot_series = None if plot_series is None else [ii.lower() for ii in plot_series]
    gather_extras = []
    allow_extras = [] if allow_extras is None else allow_extras
    marker_id = 0
    for (name, is_deploy), group in dd.groupby([dd["series"], ["deploy" in ii for ii in dd["model"]]]):
        if plot_series is not None and name.lower() not in plot_series:
            continue

        if has_extra and allow_extras != "all":
            gather_extras.extend(group["extra"].values)
            extra_condition = group["extra"].isnull()
            if allow_extras:
                extra_condition = np.logical_or(extra_condition, [ii in allow_extras for ii in group["extra"]])
            group = group[extra_condition]
        xx = group[x_label].values
        yy = group[y_label].values
        if len(xx) == 0 or len(yy) == 0:
            print("Empty or all filtered for series", name)
            continue

        ax.plot(xx, yy)
        label = (name + "_deploy") if is_deploy else name
        prefix_len = len(name.split("_")[0])
        ax.scatter(xx, yy, label=label, marker=markers[marker_id])
        marker_id = (marker_id + 1) if marker_id < len(markers) - 1 else 0
        for _, cur in group.iterrows():
            # print(cur)
            text = cur["model"][prefix_len:]
            if has_extra and str(cur["extra"]) != "nan":
                text += "," + cur["extra"]
            ax.text(cur[x_label], cur[y_label], text[1:] if text.startswith("_") else text, fontsize=fontsize)
    if log_scale_x:
        ax.set_xscale("log")
        min_value_x, max_value_x = ax.get_xlim()
        # print(f"{min_value_x = }, {max_value_x = }")
        min_value_x_log, max_value_x_log = np.log(max(min_value_x, 1e-3)) / np.log(10), np.log(max_value_x) / np.log(10)
        ticks = [10**ii for ii in np.arange(min_value_x_log, max_value_x_log, (max_value_x_log - min_value_x_log) / 10)]
        ax.set_xticks(ticks, labels=["{:.3f}".format(ii) for ii in ticks], fontsize=fontsize)

    ax.set_xlabel(x_label + " (log distribution)" if log_scale_x else "", fontsize=fontsize + 1)
    ax.set_ylabel(y_label, fontsize=fontsize + 1)
    ax.legend(fontsize=fontsize)
    ax.grid(True)
    matplotlib.rcParams["xtick.labelsize"], matplotlib.rcParams["ytick.labelsize"] = pre_x_labelsize, pre_y_labelsize

    plt.tight_layout()
    plt.show()
    other_extras = [ii for ii in set(gather_extras) - set(allow_extras) if isinstance(ii, str)]
    if len(other_extras) > 0:
        print("All other extras:", other_extras)
    return ax
